"""Add request data to tasks

Revision ID: 31255a412d63
Revises: 80a29f64f91c
Create Date: 2024-12-09 16:40:20.926138

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "31255a412d63"
down_revision: Union[str, None] = "80a29f64f91c"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    with op.batch_alter_table("tasks", schema=None) as batch_op:
        batch_op.add_column(sa.Column("request_data", sa.String(), nullable=True))


def downgrade() -> None:
    with op.batch_alter_table("tasks", schema=None) as batch_op:
        batch_op.drop_column("request_data")



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
# pylint: disable-msg=invalid-name,redefined-builtin
from os import environ

import argos

project = "Argos monitoring"
copyright = "2023, Alexis Métaireau, Framasoft"
author = "Alexis Métaireau, Framasoft"
release = argos.VERSION

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["myst_parser", "sphinx_design", "sphinxcontrib.mermaid"]
myst_enable_extensions = ["colon_fence"]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
mermaid_params = ["--theme", "forest"]

html_sidebars = {
    "**": [
        "sidebars/localtoc.html",
        "repository.html",
    ]
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

smartquotes = False

if "CI_JOB_ID" in environ:
    html_baseurl = "https://argos-monitoring.framasoft.org"

html_theme = "shibuya"
html_static_path = ["_static"]
html_css_files = ["fonts.css", "fix-nav.css"]
html_logo = "_static/logo.png"
html_theme_options = {
    "og_image_url": "https://argos-monitoring.framasoft.org/_static/logo.png"
}



"""Pydantic schemas for data

For database models, see argos.server.models.
"""
import traceback
from datetime import datetime
from typing import Literal

from pydantic import BaseModel, ConfigDict

from argos.schemas.utils import IPVersion, Method

# XXX Refactor using SQLModel to avoid duplication of model data


class Task(BaseModel):
    """A task corresponds to a check to execute"""

    id: int
    url: str
    domain: str
    ip_version: IPVersion
    check: str
    method: Method
    request_data: str | None
    expected: str
    task_group: str
    retry_before_notification: int
    contiguous_failures: int
    selected_at: datetime | None
    selected_by: str | None

    model_config = ConfigDict(from_attributes=True)

    def __str__(self):
        task_id = self.id
        url = self.url
        check = self.check
        ip_version = self.ip_version
        return f"Task ({task_id}): {url} (IPv{ip_version}) - {check}"


class SerializableException(BaseModel):
    """Task exception"""

    error_message: str
    error_type: str
    error_details: str

    @staticmethod
    def from_exception(err: BaseException):
        return SerializableException(
            error_message=str(err),
            error_type=str(type(err).__name__),
            error_details=traceback.format_exc(),
        )


class AgentResult(BaseModel):
    """Task’s result sent by agent"""

    task_id: int
    # The on-check status means that the service needs to finish the check
    # and will then determine the severity.
    status: Literal["success", "failure", "error", "on-check"]
    context: dict | SerializableException



"""Initial migrations

Revision ID: 7d480e6f1112
Revises: 
Create Date: 2023-12-16 23:33:40.059077

"""
from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "7d480e6f1112"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.create_table(
        "tasks",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("url", sa.String(), nullable=False),
        sa.Column("domain", sa.String(), nullable=False),
        sa.Column("check", sa.String(), nullable=False),
        sa.Column("expected", sa.String(), nullable=False),
        sa.Column("frequency", sa.Integer(), nullable=False),
        sa.Column("selected_by", sa.String(), nullable=True),
        sa.Column("selected_at", sa.DateTime(), nullable=True),
        sa.Column("completed_at", sa.DateTime(), nullable=True),
        sa.Column("next_run", sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "results",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("task_id", sa.Integer(), nullable=False),
        sa.Column("agent_id", sa.String(), nullable=True),
        sa.Column("submitted_at", sa.DateTime(), nullable=False),
        sa.Column(
            "status",
            sa.Enum("success", "failure", "error", "on-check", name="status"),
            nullable=False,
        ),
        sa.Column(
            "severity",
            sa.Enum("ok", "warning", "critical", name="severity"),
            nullable=False,
        ),
        sa.Column("context", sa.JSON(), nullable=False),
        sa.ForeignKeyConstraint(
            ["task_id"],
            ["tasks.id"],
            name="results_task_id_fkey",
        ),
        sa.PrimaryKeyConstraint("id"),
    )


def downgrade() -> None:
    op.drop_table("results")
    op.drop_table("tasks")



from fastapi import Request
from fastapi.responses import RedirectResponse


class NotAuthenticatedException(Exception):
    pass


def auth_exception_handler(request: Request, exc: NotAuthenticatedException):
    """
    Redirect the user to the login page if not logged in
    """
    response = RedirectResponse(url=request.url_for("login_view"))
    manager = request.app.state.manager
    manager.set_cookie(response, "")
    return response



from datetime import datetime, timedelta

import pytest

from argos import schemas
from argos.server import queries
from argos.server.models import Result, Task, User


@pytest.mark.asyncio
async def test_remove_old_results(db, ten_tasks):  #  pylint: disable-msg=redefined-outer-name
    for _task in ten_tasks:
        for _ in range(5):
            result = Result(
                submitted_at=datetime.now(),
                status="success",
                context={"foo": "bar"},
                task=_task,
                agent_id="test",
                severity="ok",
            )
            db.add(result)
    db.commit()

    # So we have 5 results per tasks
    assert db.query(Result).count() == 50
    # Keep only 2
    deleted = await queries.remove_old_results(db, 2)
    assert deleted == 30
    assert db.query(Result).count() == 20
    for _task in ten_tasks:
        assert db.query(Result).filter(Result.task == _task).count() == 2


@pytest.mark.asyncio
async def test_remove_old_results_with_empty_db(db):
    assert db.query(Result).count() == 0
    deleted = await queries.remove_old_results(db, 2)
    assert deleted == 0


@pytest.mark.asyncio
async def test_release_old_locks(db, ten_locked_tasks, ten_tasks):  #  pylint: disable-msg=redefined-outer-name
    assert db.query(Task).count() == 20
    released = await queries.release_old_locks(db, 10)
    assert released == 10


@pytest.mark.asyncio
async def test_release_old_locks_with_empty_db(db):
    assert db.query(Task).count() == 0
    released = await queries.release_old_locks(db, 10)
    assert released == 0


@pytest.mark.asyncio
async def test_update_from_config_with_duplicate_tasks(db, empty_config):  #  pylint: disable-msg=redefined-outer-name
    # We pass the same path twice
    fake_path = {"path": "/", "checks": [{"body-contains": "foo"}]}
    website = schemas.config.Website(
        domain="https://example.org",
        paths=[
            fake_path,
            fake_path,
        ],
    )
    empty_config.websites = [website]

    assert db.query(Task).count() == 0
    await queries.update_from_config(db, empty_config)

    # Only one path has been saved in the database
    assert db.query(Task).count() == 2

    # Calling again with the same data works, and will not result in more tasks being
    # created.
    await queries.update_from_config(db, empty_config)


@pytest.mark.asyncio
async def test_update_from_config_db_can_remove_duplicates_and_old_tasks(
    db,
    empty_config,
    task,  #  pylint: disable-msg=redefined-outer-name
):
    # Add a duplicate in the db
    same_task = Task(
        url=task.url,
        domain=task.domain,
        ip_version="6",
        check=task.check,
        expected=task.expected,
        frequency=task.frequency,
    )
    db.add(same_task)
    db.commit()
    assert db.query(Task).count() == 2

    website = schemas.config.Website(
        domain=task.domain,
        paths=[
            {
                "path": "https://another-example.com",
                "checks": [{task.check: task.expected}],
            },
            {"path": task.url, "checks": [{task.check: task.expected}]},
        ],
    )
    empty_config.websites = [website]

    await queries.update_from_config(db, empty_config)
    assert db.query(Task).count() == 4

    website = schemas.config.Website(
        domain=task.domain,
        paths=[
            {
                "path": "https://another-example.com",
                "checks": [{task.check: task.expected}],
            }
        ],
    )
    empty_config.websites = [website]

    await queries.update_from_config(db, empty_config)
    assert db.query(Task).count() == 2


@pytest.mark.asyncio
async def test_update_from_config_db_updates_existing_tasks(db, empty_config, task):  #  pylint: disable-msg=redefined-outer-name
    assert db.query(Task).count() == 1

    website = schemas.config.Website(
        domain=task.domain,
        paths=[{"path": task.url, "checks": [{task.check: task.expected}]}],
    )
    empty_config.websites = [website]

    await queries.update_from_config(db, empty_config)
    assert db.query(Task).count() == 2


@pytest.mark.asyncio
async def test_reschedule_all(
    db,
    ten_tasks,
    ten_warning_tasks,
    ten_critical_tasks,
    ten_ok_tasks,  #  pylint: disable-msg=redefined-outer-name
):
    assert db.query(Task).count() == 40
    assert db.query(Task).filter(Task.severity == "unknown").count() == 10
    assert db.query(Task).filter(Task.severity == "warning").count() == 10
    assert db.query(Task).filter(Task.severity == "critical").count() == 10
    assert db.query(Task).filter(Task.severity == "ok").count() == 10

    one_hour_ago = datetime.now() - timedelta(hours=1)
    assert db.query(Task).filter(Task.next_run <= one_hour_ago).count() == 0

    await queries.reschedule_all(db)
    assert db.query(Task).filter(Task.next_run <= one_hour_ago).count() == 30


@pytest.mark.asyncio
async def test_add_user(db):
    users = await queries.list_users(db)
    assert users.count() == 0

    _user = await queries.add_user(db, "john", "doe")
    assert _user.username == "john"
    assert _user.password == "doe"
    assert _user.disabled == False
    assert _user.created_at is not None
    assert _user.updated_at is None
    assert _user.last_login_at is None

    _user = await queries.get_user(db, "morgan")
    assert _user is None

    _user = await queries.get_user(db, "john")
    assert _user.username == "john"
    assert _user.password == "doe"
    assert _user.disabled == False
    assert _user.created_at is not None
    assert _user.updated_at is None
    assert _user.last_login_at is None

    users = await queries.list_users(db)
    assert users.count() == 1


@pytest.mark.asyncio
async def test_remove_user(db, user):  #  pylint: disable-msg=redefined-outer-name
    users = await queries.list_users(db)
    assert users.count() == 1

    assert user.username == "jane"
    assert user.password == "doe"
    assert user.disabled == False
    assert user.created_at is not None
    assert user.updated_at is None
    assert user.last_login_at is None

    db.delete(user)
    db.commit()

    users = await queries.list_users(db)
    assert users.count() == 0


@pytest.fixture
def task(db):
    _task = Task(
        url="https://www.example.com",
        domain="https://www.example.com",
        ip_version="6",
        check="body-contains",
        expected="foo",
        frequency=1,
    )
    db.add(_task)
    db.commit()
    return _task


@pytest.fixture
def empty_config():
    return schemas.config.Config(
        general=schemas.config.General(
            db=schemas.config.DbSettings(url="sqlite:////tmp/test-argos.db"),
            cookie_secret="foo-bar-baz",
            frequency="1m",
            alerts=schemas.config.Alert(
                ok=["", ""],
                warning=["", ""],
                critical=["", ""],
                unknown=["", ""],
            ),
        ),
        service=schemas.config.Service(
            secrets=[
                "1234",
            ]
        ),
        ssl=schemas.config.SSL(thresholds=[]),
        websites=[],
    )


@pytest.fixture
def ten_results(db, task):  #  pylint: disable-msg=redefined-outer-name
    results = []
    for _ in range(10):
        result = Result(
            submitted_at=datetime.now(),
            status="success",
            context={"foo": "bar"},
            task=task,
            agent_id="test",
            severity="ok",
        )
        db.add(result)
        results.append(result)
    db.commit()
    return results


@pytest.fixture
def ten_locked_tasks(db):
    a_minute_ago = datetime.now() - timedelta(minutes=1)
    tasks = []
    for _ in range(10):
        _task = Task(
            url="https://www.example.com",
            domain="example.com",
            ip_version="6",
            check="body-contains",
            expected="foo",
            frequency=1,
            selected_by="test",
            selected_at=a_minute_ago,
        )
        db.add(_task)
        tasks.append(_task)
    db.commit()
    return tasks


@pytest.fixture
def ten_tasks(db):
    now = datetime.now()
    tasks = []
    for _ in range(10):
        _task = Task(
            url="https://www.example.com",
            domain="example.com",
            ip_version="6",
            check="body-contains",
            expected="foo",
            frequency=1,
            selected_by="test",
            selected_at=now,
        )
        db.add(_task)
        tasks.append(_task)
    db.commit()
    return tasks


@pytest.fixture
def ten_warning_tasks(db):
    now = datetime.now()
    tasks = []
    for _ in range(10):
        _task = Task(
            url="https://www.example.com",
            domain="example.com",
            ip_version="6",
            check="body-contains",
            expected="foo",
            frequency=1,
            next_run=now,
            severity="warning",
        )
        db.add(_task)
        tasks.append(_task)
    db.commit()
    return tasks


@pytest.fixture
def ten_critical_tasks(db):
    now = datetime.now()
    tasks = []
    for _ in range(10):
        _task = Task(
            url="https://www.example.com",
            domain="example.com",
            ip_version="6",
            check="body-contains",
            expected="foo",
            frequency=1,
            next_run=now,
            severity="critical",
        )
        db.add(_task)
        tasks.append(_task)
    db.commit()
    return tasks


@pytest.fixture
def ten_ok_tasks(db):
    now = datetime.now()
    tasks = []
    for _ in range(10):
        _task = Task(
            url="https://www.example.com",
            domain="example.com",
            ip_version="6",
            check="body-contains",
            expected="foo",
            frequency=1,
            next_run=now,
            severity="ok",
        )
        db.add(_task)
        tasks.append(_task)
    db.commit()
    return tasks


@pytest.fixture
def user(db):
    _user = User(
        username="jane",
        password="doe",
        disabled=False,
    )
    db.add(_user)
    db.commit()
    return _user



import asyncio
import os

import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient
from sqlalchemy.orm import Session

os.environ["ARGOS_YAML_FILE"] = "tests/config.yaml"


@pytest.fixture
def db() -> Session:  # type: ignore[misc]
    from argos.server import models

    app = _create_app()
    models.Base.metadata.create_all(bind=app.state.engine)
    yield app.state.SessionLocal()
    models.Base.metadata.drop_all(bind=app.state.engine)


@pytest.fixture
def app() -> FastAPI:  # type: ignore[misc]
    from argos.server import models

    app = _create_app()
    models.Base.metadata.create_all(bind=app.state.engine)
    yield app
    models.Base.metadata.drop_all(bind=app.state.engine)


@pytest.fixture
def authorized_client(app):
    with TestClient(app) as client:
        token = app.state.config.service.secrets[0]
        client.headers = {"Authorization": f"Bearer {token}"}
        yield client


def _create_app() -> FastAPI:
    from argos.server.main import (  # local import for testing purpose
        get_application,
        setup_database,
        connect_to_db,
    )

    app = get_application()

    setup_database(app)
    asyncio.run(connect_to_db(app))
    return app



"""Specify check method

Revision ID: dcf73fa19fce
Revises: c780864dc407
Create Date: 2024-11-26 14:40:27.510587

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "dcf73fa19fce"
down_revision: Union[str, None] = "c780864dc407"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    with op.batch_alter_table("tasks", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "method",
                sa.Enum(
                    "GET",
                    "HEAD",
                    "POST",
                    "OPTIONS",
                    "CONNECT",
                    "TRACE",
                    "PUT",
                    "PATCH",
                    "DELETE",
                    name="method",
                ),
                nullable=False,
                server_default="GET",
            )
        )


def downgrade() -> None:
    with op.batch_alter_table("tasks", schema=None) as batch_op:
        batch_op.drop_column("method")



"""Web interface for humans"""
from collections import defaultdict
from datetime import datetime, timedelta
from functools import cmp_to_key
from pathlib import Path
from typing import Annotated
from urllib.parse import urlparse

from fastapi import APIRouter, Cookie, Depends, Form, Request, status
from fastapi.responses import RedirectResponse
from fastapi.security import OAuth2PasswordRequestForm
from fastapi.templating import Jinja2Templates
from passlib.context import CryptContext
from sqlalchemy import func
from sqlalchemy.orm import Session

from argos.checks.base import Status
from argos.schemas import Config
from argos.server import queries
from argos.server.exceptions import NotAuthenticatedException
from argos.server.models import Result, Task, User
from argos.server.routes.dependencies import get_config, get_db, get_manager

route = APIRouter()

current_dir = Path(__file__).resolve().parent
templates = Jinja2Templates(directory=current_dir / ".." / "templates")
SEVERITY_LEVELS = {"ok": 1, "warning": 2, "critical": 3, "unknown": 4}


@route.get("/login")
async def login_view(
    request: Request,
    msg: str | None = None,
    config: Config = Depends(get_config),
):
    if config.general.unauthenticated_access == "all":
        return RedirectResponse(
            request.url_for("get_severity_counts_view"),
            status_code=status.HTTP_303_SEE_OTHER,
        )

    token = request.cookies.get("access-token")
    if token is not None and token != "":
        manager = request.app.state.manager
        user = await manager.get_current_user(token)
        if user is not None:
            return RedirectResponse(
                request.url_for("get_severity_counts_view"),
                status_code=status.HTTP_303_SEE_OTHER,
            )

    if msg == "logout":
        msg = "You have been successfully disconnected."
    else:
        msg = None

    return templates.TemplateResponse(
        "login.html",
        {
            "request": request,
            "msg": msg,
            "remember": config.general.remember_me_duration,
        },
    )


@route.post("/login")
async def post_login(
    request: Request,
    db: Session = Depends(get_db),
    data: OAuth2PasswordRequestForm = Depends(),
    rememberme: Annotated[str | None, Form()] = None,
    config: Config = Depends(get_config),
):
    if config.general.unauthenticated_access == "all":
        return RedirectResponse(
            request.url_for("get_severity_counts_view"),
            status_code=status.HTTP_303_SEE_OTHER,
        )

    username = data.username

    invalid_credentials = templates.TemplateResponse(
        "login.html",
        {"request": request, "msg": "Sorry, invalid username or bad password."},
    )

    if config.general.ldap is not None:
        from ldap import INVALID_CREDENTIALS  # pylint: disable-msg=no-name-in-module
        from argos.server.routes.dependencies import find_ldap_user

        ldap_dn = await find_ldap_user(config, request.app.state.ldap, username)
        if ldap_dn is None:
            return invalid_credentials
        try:
            request.app.state.ldap.simple_bind_s(ldap_dn, data.password)
        except INVALID_CREDENTIALS:
            return invalid_credentials
    else:
        user = await queries.get_user(db, username)
        if user is None:
            return invalid_credentials

        pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")
        if not pwd_context.verify(data.password, user.password):
            return invalid_credentials

        user.last_login_at = datetime.now()
        db.commit()

    manager = request.app.state.manager
    session_duration = config.general.session_duration
    if config.general.remember_me_duration is not None and rememberme == "on":
        session_duration = config.general.remember_me_duration
    delta = timedelta(minutes=session_duration)
    token = manager.create_access_token(data={"sub": username}, expires=delta)
    response = RedirectResponse(
        request.url_for("get_severity_counts_view"),
        status_code=status.HTTP_303_SEE_OTHER,
    )
    response.set_cookie(
        key=manager.cookie_name,
        value=token,
        httponly=True,
        samesite="strict",
        expires=int(delta.total_seconds()),
    )
    return response


@route.get("/logout")
async def logout_view(
    request: Request,
    config: Config = Depends(get_config),
    user: User | None = Depends(get_manager),
):
    if config.general.unauthenticated_access == "all":
        return RedirectResponse(
            request.url_for("get_severity_counts_view"),
            status_code=status.HTTP_303_SEE_OTHER,
        )

    response = RedirectResponse(
        request.url_for("login_view").include_query_params(msg="logout"),
        status_code=status.HTTP_303_SEE_OTHER,
    )
    response.delete_cookie(key="access-token")
    return response


@route.get("/")
async def get_severity_counts_view(
    request: Request,
    user: User | None = Depends(get_manager),
    db: Session = Depends(get_db),
    auto_refresh_enabled: Annotated[bool, Cookie()] = False,
    auto_refresh_seconds: Annotated[int, Cookie()] = 15,
):
    """Shows the number of results per severity"""
    counts_dict = await queries.get_severity_counts(db)

    agents = db.query(Result.agent_id).distinct().all()

    return templates.TemplateResponse(
        "index.html",
        {
            "request": request,
            "counts_dict": counts_dict,
            "agents": agents,
            "auto_refresh_enabled": auto_refresh_enabled,
            "auto_refresh_seconds": auto_refresh_seconds,
            "user": user,
        },
    )


@route.get("/domains")
async def get_domains_view(
    request: Request,
    user: User | None = Depends(get_manager),
    config: Config = Depends(get_config),
    db: Session = Depends(get_db),
):
    """Show all tasks and their current state"""
    if config.general.unauthenticated_access == "dashboard":
        if user is None:
            raise NotAuthenticatedException

    tasks = db.query(Task).all()

    domains_severities = defaultdict(list)
    domains_last_checks = defaultdict(list)  # type: ignore[var-annotated]

    for task in tasks:
        domain = urlparse(task.url).netloc
        domains_severities[domain].append(task.severity)
        if task.last_severity_update is not None:
            domains_last_checks[domain] = task.last_severity_update
        else:
            domains_last_checks[domain] = "Waiting to be checked"

    def _max_severity(severities):
        return max(severities, key=SEVERITY_LEVELS.get)

    def _cmp_domains(a, b):
        if SEVERITY_LEVELS[a[1]] < SEVERITY_LEVELS[b[1]]:
            return 1
        if SEVERITY_LEVELS[a[1]] > SEVERITY_LEVELS[b[1]]:
            return -1
        if a[0] > b[0]:
            return 1
        if a[0] < b[0]:
            return -1
        return 0

    domains = [(key, _max_severity(value)) for key, value in domains_severities.items()]
    domains.sort(key=cmp_to_key(_cmp_domains))

    agents = db.query(Result.agent_id).distinct().all()

    return templates.TemplateResponse(
        "domains.html",
        {
            "request": request,
            "domains": domains,
            "last_checks": domains_last_checks,
            "total_task_count": len(tasks),
            "agents": agents,
            "user": user,
        },
    )


@route.get("/domain/{domain}")
async def get_domain_tasks_view(
    request: Request,
    domain: str,
    user: User | None = Depends(get_manager),
    config: Config = Depends(get_config),
    db: Session = Depends(get_db),
):
    """Show all tasks attached to a domain"""
    if config.general.unauthenticated_access == "dashboard":
        if user is None:
            raise NotAuthenticatedException

    tasks = db.query(Task).filter(Task.domain.contains(f"//{domain}")).all()
    return templates.TemplateResponse(
        "domain.html",
        {
            "request": request,
            "domain": domain,
            "tasks": tasks,
            "user": user,
        },
    )


@route.get("/result/{result_id}")
async def get_result_view(
    request: Request,
    result_id: int,
    user: User | None = Depends(get_manager),
    config: Config = Depends(get_config),
    db: Session = Depends(get_db),
):
    """Show the details of a result"""
    if config.general.unauthenticated_access == "dashboard":
        if user is None:
            raise NotAuthenticatedException

    result = db.query(Result).get(result_id)
    return templates.TemplateResponse(
        "result.html",
        {
            "request": request,
            "result": result,
            "error": Status.ERROR,
            "user": user,
        },
    )


@route.get("/task/{task_id}/results")
async def get_task_results_view(
    request: Request,
    task_id: int,
    user: User | None = Depends(get_manager),
    db: Session = Depends(get_db),
    config: Config = Depends(get_config),
):
    """Show history of a task’s results"""
    if config.general.unauthenticated_access == "dashboard":
        if user is None:
            raise NotAuthenticatedException

    results = (
        db.query(Result)
        .filter(Result.task_id == task_id)
        .order_by(Result.submitted_at.desc())  # type: ignore[attr-defined]
        .all()
    )
    task = db.query(Task).get(task_id)
    description = ""
    if task is not None:
        description = task.get_check().get_description(config)
    return templates.TemplateResponse(
        "results.html",
        {
            "request": request,
            "results": results,
            "task": task,
            "description": description,
            "error": Status.ERROR,
            "user": user,
        },
    )


@route.get("/agents")
async def get_agents_view(
    request: Request,
    user: User | None = Depends(get_manager),
    config: Config = Depends(get_config),
    db: Session = Depends(get_db),
):
    """Show argos agents and the last time the server saw them"""
    if config.general.unauthenticated_access == "dashboard":
        if user is None:
            raise NotAuthenticatedException

    last_seen = (
        db.query(Result.agent_id, func.max(Result.submitted_at).label("submitted_at"))
        .group_by(Result.agent_id)
        .all()
    )

    return templates.TemplateResponse(
        "agents.html",
        {
            "request": request,
            "last_seen": last_seen,
            "user": user,
        },
    )


@route.post("/refresh")
async def set_refresh_cookies_view(
    request: Request,
    user: User | None = Depends(get_manager),
    auto_refresh_enabled: Annotated[bool, Form()] = False,
    auto_refresh_seconds: Annotated[int, Form()] = 15,
):
    response = RedirectResponse(
        request.url_for("get_severity_counts_view"),
        status_code=status.HTTP_303_SEE_OTHER,
    )
    # Cookies’ age in Chrome can’t be more than 400 days
    # https://developer.chrome.com/blog/cookie-max-age-expires
    delta = int(timedelta(days=400).total_seconds())
    response.set_cookie(
        key="auto_refresh_enabled",
        value=str(auto_refresh_enabled),
        httponly=True,
        samesite="strict",
        expires=delta,
    )
    response.set_cookie(
        key="auto_refresh_seconds",
        value=str(max(5, int(auto_refresh_seconds))),
        httponly=True,
        samesite="strict",
        expires=delta,
    )
    return response



"""Add severity to Task and add severity level UNKNOWN

Revision ID: e99bc35702c9
Revises: 7d480e6f1112
Create Date: 2024-02-28 14:14:22.519918

"""
from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision: str = "e99bc35702c9"
down_revision: Union[str, None] = "7d480e6f1112"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    bind = op.get_bind()
    if bind.engine.name != "sqlite":
        op.execute("ALTER TYPE severity ADD VALUE 'unknown'")
    op.add_column(
        "tasks",
        sa.Column(
            "severity",
            sa.Enum("ok", "warning", "critical", "unknown", name="severity"),
            nullable=False,
        ),
    )
    op.add_column(
        "tasks", sa.Column("last_severity_update", sa.DateTime(), nullable=True)
    )


def downgrade() -> None:
    op.drop_column("tasks", "last_severity_update")
    op.drop_column("tasks", "severity")



import asyncio

import pytest
from fastapi.testclient import TestClient

from argos.schemas import AgentResult, SerializableException
from argos.server import models
from argos.server.queries import update_from_config


def test_read_tasks_requires_auth(app):
    with TestClient(app) as client:
        response = client.get("/api/tasks")
        assert response.status_code == 403


def test_tasks_retrieval_and_results(authorized_client, app):
    asyncio.run(update_from_config(app.state.db, app.state.config))
    with authorized_client as client:
        response = client.get("/api/tasks")
        assert response.status_code == 200

        tasks = response.json()
        assert len(tasks) == 4

        results = []
        for task in tasks:
            results.append(
                AgentResult(task_id=task["id"], status="success", context={})
            )

        data = [r.model_dump() for r in results]
        response = client.post("/api/results", json=data)

        assert response.status_code == 201
        assert app.state.db.query(models.Result).count() == 4

        # The list of tasks should be empty now
        response = client.get("/api/tasks")
        assert len(response.json()) == 0


def test_agents_can_report_errors(authorized_client):
    with authorized_client as client:
        exc = Exception("This is an error")
        serialized_exc = SerializableException.from_exception(exc)
        agent_result = AgentResult(task_id=1, status="error", context=serialized_exc)

        response = client.post(
            "/api/results",
            json=[
                agent_result.model_dump(),
            ],
        )
        assert response.status_code == 201


@pytest.fixture
def ssl_task(db):
    task = models.Task(
        url="https://exemple.com/",
        domain="https://exemple.com/",
        ip_version="6",
        method="GET",
        check="ssl-certificate-expiration",
        expected="on-check",
        frequency=1,
    )
    db.add(task)
    db.commit()
    return task


def test_specialized_checks_can_report_errors(authorized_client, ssl_task):
    with authorized_client as client:
        exc = Exception("This is an error")
        serialized_exc = SerializableException.from_exception(exc)
        agent_result = AgentResult(
            task_id=ssl_task.id, status="error", context=serialized_exc
        )

        response = client.post(
            "/api/results",
            json=[
                agent_result.model_dump(),
            ],
        )
        assert response.status_code == 201



"""Use bigint for results id field

Revision ID: bd4b4962696a
Revises: 31255a412d63
Create Date: 2025-01-06 11:44:37.552965

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "bd4b4962696a"
down_revision: Union[str, None] = "31255a412d63"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    bind = op.get_bind()
    if bind.engine.name != "sqlite":
        with op.batch_alter_table("results", schema=None) as batch_op:
            batch_op.alter_column(
                "id",
                existing_type=sa.INTEGER(),
                type_=sa.BigInteger(),
                existing_nullable=False,
            )


def downgrade() -> None:
    bind = op.get_bind()
    if bind.engine.name != "sqlite":
        with op.batch_alter_table("results", schema=None) as batch_op:
            batch_op.alter_column(
                "id",
                existing_type=sa.BigInteger(),
                type_=sa.INTEGER(),
                existing_nullable=False,
            )



import pytest

from argos.schemas.config import SSL, WebsitePath


def test_ssl_duration_parsing():
    data = {"thresholds": [{"2d": "warning"}, {"3w": "error"}]}

    # Test the validation and parsing of SSL model
    ssl_object = SSL(**data)
    assert len(ssl_object.thresholds) == 2
    assert ssl_object.thresholds == [(2, "warning"), (21, "error")]

    # Test the constraint on severity
    with pytest.raises(ValueError):
        erroneous_data = {"thresholds": [{"1d": "caution"}, {"1w": "danger"}]}
        SSL(**erroneous_data)


def test_path_parsing_transforms_to_tuples():
    data = {"path": "/", "checks": [{"body-contains": "youpi"}, {"status-is": "200"}]}
    path = WebsitePath(**data)
    assert len(path.checks) == 2
    assert path.checks == [("body-contains", "youpi"), ("status-is", "200")]


def test_path_ensures_check_exists():
    with pytest.raises(ValueError):
        erroneous_data = {
            "path": "/",
            "checks": [{"non-existing-key": "youpi"}, {"status-is": "200"}],
        }
        WebsitePath(**erroneous_data)


def test_expected_accepts_and_convert_ints():
    data = {"path": "/", "checks": [{"body-contains": "youpi"}, {"status-is": 200}]}
    path = WebsitePath(**data)
    assert len(path.checks) == 2
    assert path.checks == [("body-contains", "youpi"), ("status-is", "200")]



"""Add retries before notification feature

Revision ID: 80a29f64f91c
Revises: 8b58ced14d6e
Create Date: 2024-12-04 17:03:35.104368

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "80a29f64f91c"
down_revision: Union[str, None] = "8b58ced14d6e"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    with op.batch_alter_table("tasks", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "retry_before_notification",
                sa.Integer(),
                server_default="0",
                nullable=False,
            )
        )
        batch_op.add_column(
            sa.Column(
                "contiguous_failures", sa.Integer(), server_default="0", nullable=False
            )
        )


def downgrade() -> None:
    with op.batch_alter_table("tasks", schema=None) as batch_op:
        batch_op.drop_column("contiguous_failures")
        batch_op.drop_column("retry_before_notification")



"""Make frequency a float

Revision ID: a1e98cf72a5c
Revises: 127d74c770bb
Create Date: 2024-11-27 16:10:13.000705

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "a1e98cf72a5c"
down_revision: Union[str, None] = "127d74c770bb"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    with op.batch_alter_table("tasks", schema=None) as batch_op:
        batch_op.alter_column(
            "frequency",
            existing_type=sa.INTEGER(),
            type_=sa.Float(),
            existing_nullable=False,
        )


def downgrade() -> None:
    with op.batch_alter_table("tasks", schema=None) as batch_op:
        batch_op.alter_column(
            "frequency",
            existing_type=sa.Float(),
            type_=sa.INTEGER(),
            existing_nullable=False,
        )



"""Add users table

Revision ID: c780864dc407
Revises: defda3f2952d
Create Date: 2024-06-10 16:31:17.296983

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "c780864dc407"
down_revision: Union[str, None] = "defda3f2952d"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.create_table(
        "users",
        sa.Column("username", sa.String(), nullable=False),
        sa.Column("password", sa.String(), nullable=False),
        sa.Column("disabled", sa.Boolean(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.Column("last_login_at", sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint("username"),
    )


def downgrade() -> None:
    op.drop_table("users")



import datetime
import ssl
from unittest.mock import MagicMock

import httpx
import pytest

from argos.checks import SSLCertificateExpiration
from argos.schemas import Task


@pytest.fixture
def now():
    return datetime.datetime.now()


@pytest.fixture
def httpx_extensions_ssl():
    """Returns the httpx extension dict used by the SSL verification check,
    to be used when mocking the client responses"""
    ssl_context = ssl.create_default_context()
    ssl_context.check_hostname = False
    ssl_context.verify_mode = ssl.CERT_NONE

    ssl_obj = MagicMock()
    ssl_obj.getpeercert.return_value = {"notAfter": "Jan 25 20:35:00 2024 GMT"}
    network_stream = MagicMock()
    network_stream.get_extra_info = MagicMock(return_value=ssl_obj)
    return {"network_stream": network_stream}


@pytest.fixture
def ssl_task(now):
    return Task(
        id=1,
        url="https://example.org",
        domain="https://example.org",
        ip_version="6",
        method="GET",
        request_data=None,
        task_group="GET-6-https://example.org",
        check="ssl-certificate-expiration",
        retry_before_notification=0,
        contiguous_failures=0,
        expected="on-check",
        selected_at=now,
        selected_by="pytest",
    )


@pytest.mark.parametrize("http_status", ["200", "300", "400", "500"])
@pytest.mark.asyncio
async def test_ssl_check_accepts_statuts(
    respx_mock, ssl_task, httpx_extensions_ssl, http_status
):
    respx_mock.get("https://example.org").mock(
        return_value=httpx.Response(http_status, extensions=httpx_extensions_ssl),
    )
    async with httpx.AsyncClient() as client:
        check = SSLCertificateExpiration(ssl_task)
        response = await client.request(
            method=ssl_task.method, url=ssl_task.url, timeout=60
        )
        check_response = await check.run(response)
        assert check_response.status == "on-check"



import ssl
import smtplib
from email.message import EmailMessage

from typing import List
from urllib.parse import urlparse

import apprise
import httpx

from argos.checks.base import Severity
from argos.logging import logger
from argos.schemas.config import Config, Mail, GotifyUrl
from argos.server.models import Task


def need_alert(
    last_severity: str, last_severity_update, severity: str, status: str, task: Task
) -> bool:
    ## Create alert… or not!
    send_notif = False
    # Severity has changed, and no retry before notification
    if last_severity != severity and task.retry_before_notification == 0:
        send_notif = True
    # Seems to be a first check: create a notification
    elif last_severity != severity and last_severity_update is None:
        send_notif = True
        # As we created a notification, avoid resending it on a
        # future failure
        if status != "success":
            task.contiguous_failures = task.retry_before_notification
    # We need retry before notification, so the severity may not have changed
    # since last check
    elif task.retry_before_notification != 0:
        # If we got a success, and we already have created a notification:
        # create notification of success immediately
        if (
            status == "success"
            and task.contiguous_failures >= task.retry_before_notification
        ):
            send_notif = True
            task.contiguous_failures = 0
        # The status is not a success
        elif status != "success":
            # This is a new failure
            task.contiguous_failures += 1
            # Severity has changed, but not to success, that’s odd:
            # create a notification
            if (
                last_severity not in ("ok", severity)
                and last_severity_update is not None
            ):
                send_notif = True
                # As we created a notification, avoid resending it on a
                # future failure
                task.contiguous_failures = task.retry_before_notification
            # Severity has not changed, but there has been enough failures
            # to create a notification
            elif task.retry_before_notification + 1 == task.contiguous_failures:
                send_notif = True

    return send_notif


def get_icon_from_severity(severity: str) -> str:
    icon = "❌"
    if severity == Severity.OK:
        icon = "✅"
    elif severity == Severity.WARNING:
        icon = "⚠️"
    elif severity == Severity.UNKNOWN:
        icon = "❔"

    return icon


def handle_alert(config: Config, result, task, severity, old_severity, request):  # pylint: disable-msg=too-many-positional-arguments
    """Dispatch alert through configured alert channels"""

    if "local" in getattr(config.general.alerts, severity):
        logger.error(
            "Alerting stub: task=%i, status=%s, severity=%s",
            task.id,
            result.status,
            severity,
        )

    if config.general.mail is not None and "mail" in getattr(
        config.general.alerts, severity
    ):
        notify_by_mail(
            result, task, severity, old_severity, config.general.mail, request
        )

    if config.general.gotify is not None and "gotify" in getattr(
        config.general.alerts, severity
    ):
        notify_with_gotify(
            result, task, severity, old_severity, config.general.gotify, request
        )

    if config.general.apprise is not None:
        for notif_way in getattr(config.general.alerts, severity):
            if notif_way.startswith("apprise:"):
                group = notif_way[8:]
                notify_with_apprise(
                    result,
                    task,
                    severity,
                    old_severity,
                    config.general.apprise[group],
                    request,
                )


def notify_with_apprise(  # pylint: disable-msg=too-many-positional-arguments
    result, task, severity: str, old_severity: str, group: List[str], request
) -> None:
    logger.debug("Will send apprise notification")

    apobj = apprise.Apprise()
    for channel in group:
        apobj.add(channel)

    icon = get_icon_from_severity(severity)
    title = f"[Argos] {icon} {urlparse(task.url).netloc} (IPv{task.ip_version}): status {severity}"
    msg = f"""\
URL:    {task.url} (IPv{task.ip_version})
Check:  {task.check}
Status: {severity}
Time:   {result.submitted_at}
Previous status: {old_severity}

See result on {request.url_for('get_result_view', result_id=result.id)}

See results of task on {request.url_for('get_task_results_view', task_id=task.id)}#{result.id}
"""

    apobj.notify(title=title, body=msg)


def notify_by_mail(  # pylint: disable-msg=too-many-positional-arguments
    result, task, severity: str, old_severity: str, config: Mail, request
) -> None:
    logger.debug("Will send mail notification")

    icon = get_icon_from_severity(severity)
    msg = f"""\
URL:    {task.url} (IPv{task.ip_version})
Check:  {task.check}
Status: {severity}
Time:   {result.submitted_at}
Previous status: {old_severity}

See result on {request.url_for('get_result_view', result_id=result.id)}

See results of task on {request.url_for('get_task_results_view', task_id=task.id)}#{result.id}
"""

    mail = EmailMessage()
    mail[
        "Subject"
    ] = f"[Argos] {icon} {urlparse(task.url).netloc} (IPv{task.ip_version}): status {severity}"
    mail["From"] = config.mailfrom
    mail.set_content(msg)

    if config.ssl:
        logger.debug("Mail notification: SSL")
        context = ssl.create_default_context()
        smtp = smtplib.SMTP_SSL(host=config.host, port=config.port, context=context)
    else:
        smtp = smtplib.SMTP(
            host=config.host,  # type: ignore
            port=config.port,
        )
        if config.starttls:
            logger.debug("Mail notification: STARTTLS")
            context = ssl.create_default_context()
            smtp.starttls(context=context)

    if config.auth is not None:
        logger.debug("Mail notification: authentification")
        smtp.login(config.auth.login, config.auth.password)

    for address in config.addresses:
        logger.debug("Sending mail to %s", address)
        logger.debug(msg)
        smtp.send_message(mail, to_addrs=address)


def notify_with_gotify(  # pylint: disable-msg=too-many-positional-arguments
    result, task, severity: str, old_severity: str, config: List[GotifyUrl], request
) -> None:
    logger.debug("Will send gotify notification")
    headers = {"accept": "application/json", "content-type": "application/json"}

    icon = get_icon_from_severity(severity)
    priority = 9
    if severity == Severity.OK:
        priority = 1
    elif severity == Severity.WARNING:
        priority = 5
    elif severity == Severity.UNKNOWN:
        priority = 5

    subject = (
        f"{icon} {urlparse(task.url).netloc} (IPv{task.ip_version}): status {severity}"
    )
    msg = f"""\
URL:    <{task.url}> (IPv{task.ip_version})\\
Check:  {task.check}\\
Status: {severity}\\
Time:   {result.submitted_at}\\
Previous status: {old_severity}\\
\\
See result on <{request.url_for('get_result_view', result_id=result.id)}>\\
\\
See results of task on <{request.url_for('get_task_results_view', task_id=task.id)}#{result.id}>
"""
    extras = {
        "client::display": {"contentType": "text/markdown"},
        "client::notification": {
            "click": {
                "url": f"{request.url_for('get_result_view', result_id=result.id)}"
            }
        },
    }

    payload = {"title": subject, "message": msg, "priority": priority, "extras": extras}

    for url in config:
        logger.debug("Sending gotify message(s) to %s", url.url)
        for token in url.tokens:
            try:
                res = httpx.post(
                    f"{url.url}message",
                    params={"token": token},
                    headers=headers,
                    json=payload,
                )
                res.raise_for_status()
            except httpx.RequestError as err:
                logger.error(
                    "An error occurred while sending a message to %s with token %s",
                    err.request.url,
                    token,
                )



"""Web interface for machines"""
from typing import List

from fastapi import APIRouter, BackgroundTasks, Depends, Request
from sqlalchemy.orm import Session

from argos.logging import logger
from argos.schemas import AgentResult, Config, Task
from argos.server import queries
from argos.server.alerting import handle_alert, need_alert
from argos.server.routes.dependencies import get_config, get_db, verify_token

route = APIRouter()


@route.get("/tasks", response_model=list[Task], dependencies=[Depends(verify_token)])
async def read_tasks(
    request: Request,
    db: Session = Depends(get_db),
    limit: int = 10,
    agent_id: None | str = None,
):
    """Return a list of tasks to execute"""
    host = ""
    if request.client is not None:
        host = request.client.host
    agent_id = agent_id or host
    tasks = await queries.list_tasks(db, agent_id=agent_id, limit=limit)
    return tasks


@route.post("/results", status_code=201, dependencies=[Depends(verify_token)])
async def create_results(  # pylint: disable-msg=too-many-positional-arguments
    request: Request,
    results: List[AgentResult],
    background_tasks: BackgroundTasks,
    db: Session = Depends(get_db),
    config: Config = Depends(get_config),
    agent_id: None | str = None,
):
    """Get the results from the agents and store them locally.

    - Finalize the checks (some checks need the server to do some part of the validation,
      for instance because they need access to the configuration)
    - If it's an error, determine its severity ;
    - Trigger the reporting calls
    """
    host = ""
    if request.client is not None:
        host = request.client.host
    agent_id = agent_id or host
    db_results = []
    for agent_result in results:
        # XXX Maybe offload this to a queue.
        # XXX Get all the tasks at once, to limit the queries on the db
        task = await queries.get_task(db, agent_result.task_id)
        if not task:
            logger.error("Unable to find task %i", agent_result.task_id)
        else:
            last_severity = task.severity
            last_severity_update = task.last_severity_update
            result = await queries.create_result(db, agent_result, agent_id)
            check = task.get_check()
            status, severity = await check.finalize(config, result, **result.context)
            result.set_status(status, severity)
            task.set_times_severity_and_deselect(severity, result.submitted_at)

            send_notif = need_alert(
                last_severity, last_severity_update, severity, status, task
            )

            if send_notif:
                background_tasks.add_task(
                    handle_alert,
                    config,
                    result,
                    task,
                    severity,
                    last_severity,
                    request,
                )

            db_results.append(result)
    db.commit()
    return {"result_ids": [r.id for r in db_results]}


@route.post(
    "/reschedule/all",
    responses={
        200: {
            "content": {
                "application/json": {"example": {"msg": "Non OK tasks reschuled"}}
            }
        }
    },
)
async def reschedule_all(request: Request, db: Session = Depends(get_db)):
    """Reschedule checks of all non OK tasks ASAP"""
    await queries.reschedule_all(db)
    return {"msg": "Non OK tasks reschuled"}


@route.get(
    "/stats",
    responses={
        200: {
            "content": {
                "application/json": {
                    "example": {
                        "upcoming_tasks_count": 0,
                        "results_count": 1993085,
                        "selected_tasks_count": 1845,
                    }
                }
            }
        }
    },
)
async def get_stats(db: Session = Depends(get_db)):
    """Get tasks statistics"""
    return {
        "upcoming_tasks_count": await queries.count_tasks(db, selected=False),
        "results_count": await queries.count_results(db),
        "selected_tasks_count": await queries.count_tasks(db, selected=True),
    }


@route.get(
    "/severities",
    responses={
        200: {
            "content": {
                "application/json": {
                    "example": {"ok": 1541, "warning": 0, "critical": 0, "unknown": 0}
                }
            }
        }
    },
)
async def get_severity_counts(db: Session = Depends(get_db)):
    """Returns the number of results per severity"""
    return await queries.get_severity_counts(db)



from argos.checks.base import Response, Status


def test_response_failure_with_context():
    resp = Response.new(False, some="context", another=True)
    assert resp.status == Status.FAILURE
    assert resp.context == {"some": "context", "another": True}


def test_response_success():
    resp = Response.new(True)
    assert resp.status == Status.SUCCESS


def test_response_on_check_with_context():
    resp = Response.new(Status.ON_CHECK, expires_in=3)
    assert resp.status == Status.ON_CHECK
    assert resp.status == "on-check"
    assert resp.context == {"expires_in": 3}



"""Add ON DELETE CASCADE to results’ task_id

Revision ID: defda3f2952d
Revises: 1a3497f9f71b
Create Date: 2024-03-18 15:09:34.544573

"""
from typing import Sequence, Union

from alembic import op


# revision identifiers, used by Alembic.
revision: str = "defda3f2952d"
down_revision: Union[str, None] = "1a3497f9f71b"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    bind = op.get_bind()
    with op.batch_alter_table("results", schema=None) as batch_op:
        if bind.engine.name != "sqlite":
            batch_op.drop_constraint("results_task_id_fkey", type_="foreignkey")
        batch_op.create_foreign_key(
            "results_task_id_fkey", "tasks", ["task_id"], ["id"], ondelete="CASCADE"
        )


def downgrade() -> None:
    bind = op.get_bind()
    with op.batch_alter_table("results", schema=None) as batch_op:
        if bind.engine.name != "sqlite":
            batch_op.drop_constraint("results_task_id_fkey", type_="foreignkey")
        batch_op.create_foreign_key(
            "results_task_id_fkey", "tasks", ["task_id"], ["id"]
        )



import os
import sys
from contextlib import asynccontextmanager
from pathlib import Path

from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from fastapi_login import LoginManager
from pydantic import ValidationError
from sqlalchemy import create_engine, event
from sqlalchemy.orm import sessionmaker

from argos.logging import logger
from argos.server import models, routes, queries
from argos.server.exceptions import NotAuthenticatedException, auth_exception_handler
from argos.server.settings import read_yaml_config


def get_application() -> FastAPI:
    """Spawn Argos FastAPI server"""
    config_file = os.environ["ARGOS_YAML_FILE"]
    config = read_config(config_file)

    root_path = config.general.root_path

    if root_path != "":
        logger.info("Root path for Argos: %s", root_path)
        if root_path.endswith("/"):
            root_path = root_path[:-1]
            logger.info("Fixed root path for Argos: %s", root_path)

    appli = FastAPI(lifespan=lifespan, root_path=root_path)

    # Config is the argos config object (built from yaml)
    appli.state.config = config
    appli.add_exception_handler(NotAuthenticatedException, auth_exception_handler)
    appli.state.manager = create_manager(config.general.cookie_secret)

    if config.general.ldap is not None:
        import ldap

        l = ldap.initialize(config.general.ldap.uri)
        l.simple_bind_s(config.general.ldap.bind_dn, config.general.ldap.bind_pwd)
        appli.state.ldap = l

    @appli.state.manager.user_loader()
    async def query_user(user: str) -> None | str | models.User:
        """
        Get a user from the db or LDAP
        :param user: name of the user
        :return: None or the user object
        """
        if appli.state.config.general.ldap is not None:
            from argos.server.routes.dependencies import find_ldap_user

            return await find_ldap_user(appli.state.config, appli.state.ldap, user)

        return await queries.get_user(appli.state.db, user)

    appli.include_router(routes.api, prefix="/api")
    appli.include_router(routes.views)

    static_dir = Path(__file__).resolve().parent / "static"

    appli.mount("/static", StaticFiles(directory=static_dir), name="static")
    return appli


async def connect_to_db(appli):
    appli.state.db = appli.state.SessionLocal()
    return appli.state.db


def read_config(yaml_file):
    try:
        config = read_yaml_config(yaml_file)
        return config
    except ValidationError as err:
        logger.error("Errors where found while reading configuration:")
        for error in err.errors():
            logger.error("%s is %s", error["loc"], error["type"])
        sys.exit(1)


def setup_database(appli):
    config = appli.state.config
    db_url = str(config.general.db.url)
    logger.debug("Using database URL %s", db_url)
    # For sqlite, we need to add connect_args={"check_same_thread": False}
    if config.general.env == "production" and db_url.startswith("sqlite:////tmp"):
        logger.warning("Using sqlite in /tmp is not recommended for production")

    extra_settings = {}
    if config.general.db.pool_size:
        extra_settings.setdefault("pool_size", config.general.db.pool_size)

    if config.general.db.max_overflow:
        extra_settings.setdefault("max_overflow", config.general.db.max_overflow)

    engine = create_engine(db_url, **extra_settings)

    def _fk_pragma_on_connect(dbapi_con, con_record):
        dbapi_con.execute("pragma foreign_keys=ON")

    if db_url.startswith("sqlite:///"):
        event.listen(engine, "connect", _fk_pragma_on_connect)

    appli.state.SessionLocal = sessionmaker(
        autocommit=False, autoflush=False, bind=engine
    )
    appli.state.engine = engine
    models.Base.metadata.create_all(bind=engine)


def create_manager(cookie_secret: str) -> LoginManager:
    if cookie_secret == "foo_bar_baz":
        logger.warning(
            "You should change the cookie_secret secret in your configuration file."
        )
    return LoginManager(
        cookie_secret,
        "/login",
        use_cookie=True,
        use_header=False,
        not_authenticated_exception=NotAuthenticatedException,
    )


@asynccontextmanager
async def lifespan(appli):
    """Server start and stop actions

    Setup database connection then close it at shutdown.
    """
    setup_database(appli)

    db = await connect_to_db(appli)

    tasks_count = await queries.count_tasks(db)
    if tasks_count == 0:
        logger.warning(
            "There is no tasks in the database. "
            'Please launch the command "argos server reload-config"'
        )

    yield

    appli.state.db.close()


app = get_application()



"""Add recheck delay

Revision ID: 127d74c770bb
Revises: dcf73fa19fce
Create Date: 2024-11-27 16:04:58.138768

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "127d74c770bb"
down_revision: Union[str, None] = "dcf73fa19fce"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    with op.batch_alter_table("tasks", schema=None) as batch_op:
        batch_op.add_column(sa.Column("recheck_delay", sa.Float(), nullable=True))
        batch_op.add_column(
            sa.Column(
                "already_retried",
                sa.Boolean(),
                nullable=False,
                server_default=sa.sql.false(),
            )
        )


def downgrade() -> None:
    with op.batch_alter_table("tasks", schema=None) as batch_op:
        batch_op.drop_column("already_retried")
        batch_op.drop_column("recheck_delay")



"""Functions to ease SQL queries management"""
from datetime import datetime, timedelta
from hashlib import sha256
from typing import List
from urllib.parse import urljoin

from sqlalchemy import asc, desc, func, Select
from sqlalchemy.orm import Session

from argos import schemas
from argos.logging import logger
from argos.server.models import Result, Task, ConfigCache, User


async def list_tasks(db: Session, agent_id: str, limit: int = 100):
    """List tasks and mark them as selected"""
    subquery = (
        db.query(func.distinct(Task.task_group))
        .filter(
            Task.selected_by == None,  # noqa: E711
            ((Task.next_run <= datetime.now()) | (Task.next_run == None)),  # noqa: E711
        )
        .limit(limit)
        .subquery()
    )
    tasks = db.query(Task).filter(Task.task_group.in_(Select(subquery))).all()

    now = datetime.now()
    for task in tasks:
        task.selected_at = now
        task.selected_by = agent_id
    db.commit()
    return tasks


async def add_user(db: Session, name: str, password: str) -> User:
    user = User(
        username=name,
        password=password,
        disabled=False,
    )
    db.add(user)
    db.commit()
    return user


async def get_user(db: Session, username: str) -> None | User:
    return db.get(User, username)


async def list_users(db: Session):
    return db.query(User).order_by(asc(User.username))


async def get_task(db: Session, task_id: int) -> None | Task:
    return db.get(Task, task_id)


async def create_result(db: Session, agent_result: schemas.AgentResult, agent_id: str):
    result = Result(
        submitted_at=datetime.now(),
        status=agent_result.status,
        context=agent_result.context,
        task_id=agent_result.task_id,
        agent_id=agent_id,
    )
    db.add(result)
    return result


async def count_tasks(db: Session, selected: None | bool = None):
    query = db.query(Task)
    if selected is not None:
        if selected:
            query = query.filter(Task.selected_by is not None)  # type: ignore[arg-type]
        else:
            query = query.filter(Task.selected_by is None)  # type: ignore[arg-type]

    return query.count()


async def count_results(db: Session):
    return db.query(Result).count()


async def has_config_changed(db: Session, config: schemas.Config) -> bool:  # pylint: disable-msg=too-many-statements
    """Check if websites config has changed by using a hashsum and a config cache"""
    websites_hash = sha256(str(config.websites).encode()).hexdigest()
    conf_caches = db.query(ConfigCache).all()
    same_config = True
    keys = [
        "websites_hash",
        "general_frequency",
        "general_recheck_delay",
        "general_retry_before_notification",
        "general_ipv4",
        "general_ipv6",
    ]
    if conf_caches:
        for conf in conf_caches:
            keys.remove(conf.name)
            match conf.name:
                case "websites_hash":
                    if conf.val != websites_hash:
                        same_config = False
                        conf.val = websites_hash
                        conf.updated_at = datetime.now()
                case "general_frequency":
                    if conf.val != str(config.general.frequency):
                        same_config = False
                        conf.val = str(config.general.frequency)
                        conf.updated_at = datetime.now()
                case "general_recheck_delay":
                    if conf.val != str(config.general.recheck_delay):
                        same_config = False
                        conf.val = str(config.general.recheck_delay)
                        conf.updated_at = datetime.now()
                case "general_retry_before_notification":
                    if conf.val != str(config.general.retry_before_notification):
                        same_config = False
                        conf.val = str(config.general.retry_before_notification)
                        conf.updated_at = datetime.now()
                case "general_ipv4":
                    if conf.val != str(config.general.ipv4):
                        same_config = False
                        conf.val = str(config.general.ipv4)
                        conf.updated_at = datetime.now()
                case "general_ipv6":
                    if conf.val != str(config.general.ipv6):
                        same_config = False
                        conf.val = str(config.general.ipv6)
                        conf.updated_at = datetime.now()

        for i in keys:
            match i:
                case "websites_hash":
                    c = ConfigCache(
                        name="websites_hash",
                        val=websites_hash,
                        updated_at=datetime.now(),
                    )
                case "general_frequency":
                    c = ConfigCache(
                        name="general_frequency",
                        val=str(config.general.frequency),
                        updated_at=datetime.now(),
                    )
                case "general_recheck_delay":
                    c = ConfigCache(
                        name="general_recheck_delay",
                        val=str(config.general.recheck_delay),
                        updated_at=datetime.now(),
                    )
                case "general_retry_before_notification":
                    c = ConfigCache(
                        name="general_retry_before_notification",
                        val=str(config.general.retry_before_notification),
                        updated_at=datetime.now(),
                    )
                case "general_ipv4":
                    c = ConfigCache(
                        name="general_ipv4",
                        val=str(config.general.ipv4),
                        updated_at=datetime.now(),
                    )
                case "general_ipv6":
                    c = ConfigCache(
                        name="general_ipv6",
                        val=str(config.general.ipv6),
                        updated_at=datetime.now(),
                    )
            db.add(c)

        db.commit()

        if keys:
            return True

        if same_config:
            return False

    else:  # no config cache found
        web_hash = ConfigCache(
            name="websites_hash", val=websites_hash, updated_at=datetime.now()
        )
        gen_freq = ConfigCache(
            name="general_frequency",
            val=str(config.general.frequency),
            updated_at=datetime.now(),
        )
        gen_recheck = ConfigCache(
            name="general_recheck_delay",
            val=str(config.general.recheck_delay),
            updated_at=datetime.now(),
        )
        gen_retry_before_notif = ConfigCache(
            name="general_retry_before_notification",
            val=str(config.general.retry_before_notification),
            updated_at=datetime.now(),
        )
        gen_ipv4 = ConfigCache(
            name="general_ipv4",
            val=str(config.general.ipv4),
            updated_at=datetime.now(),
        )
        gen_ipv6 = ConfigCache(
            name="general_ipv6",
            val=str(config.general.ipv6),
            updated_at=datetime.now(),
        )
        db.add(web_hash)
        db.add(gen_freq)
        db.add(gen_recheck)
        db.add(gen_retry_before_notif)
        db.add(gen_ipv4)
        db.add(gen_ipv6)
        db.commit()

    return True


async def update_from_config(db: Session, config: schemas.Config):  # pylint: disable-msg=too-many-branches
    """Update tasks from config file"""
    config_changed = await has_config_changed(db, config)
    if not config_changed:
        return {"added": 0, "vanished": 0}

    max_task_id = (
        db.query(func.max(Task.id).label("max_id")).all()  #  pylint: disable-msg=not-callable
    )[0].max_id
    tasks = []
    unique_properties = []
    seen_tasks: List[int] = []
    for website in config.websites:  # pylint: disable-msg=too-many-nested-blocks
        domain = str(website.domain)
        frequency = website.frequency or config.general.frequency
        recheck_delay = website.recheck_delay or config.general.recheck_delay
        retry_before_notification = (
            website.retry_before_notification
            if website.retry_before_notification is not None
            else config.general.retry_before_notification
        )
        ipv4 = website.ipv4 if website.ipv4 is not None else config.general.ipv4
        ipv6 = website.ipv6 if website.ipv6 is not None else config.general.ipv6
        if ipv4 is False and ipv6 is False:
            logger.warning("IPv4 AND IPv6 are disabled on website %s!", domain)
            continue

        for ip_version in ["4", "6"]:
            for p in website.paths:
                url = urljoin(domain, str(p.path))
                for check_key, expected in p.checks:
                    # Check the db for already existing tasks.
                    existing_tasks = (
                        db.query(Task)
                        .filter(
                            Task.url == url,
                            Task.method == p.method,
                            Task.request_data == p.request_data,
                            Task.check == check_key,
                            Task.expected == expected,
                            Task.ip_version == ip_version,
                        )
                        .all()
                    )

                    if (ip_version == "4" and ipv4 is False) or (
                        ip_version == "6" and ipv6 is False
                    ):
                        continue

                    if existing_tasks:
                        existing_task = existing_tasks[0]

                        seen_tasks.append(existing_task.id)

                        if frequency != existing_task.frequency:
                            existing_task.frequency = frequency
                        if recheck_delay != existing_task.recheck_delay:
                            existing_task.recheck_delay = recheck_delay  # type: ignore[assignment]
                        if (
                            retry_before_notification
                            != existing_task.retry_before_notification
                        ):
                            existing_task.retry_before_notification = (
                                retry_before_notification
                            )
                        logger.debug(
                            "Skipping db task creation for url=%s, "
                            "method=%s, check_key=%s, expected=%s, "
                            "frequency=%s, recheck_delay=%s, "
                            "retry_before_notification=%s, ip_version=%s.",
                            url,
                            p.method,
                            check_key,
                            expected,
                            frequency,
                            recheck_delay,
                            retry_before_notification,
                            ip_version,
                        )

                    else:
                        properties = (
                            url,
                            p.method,
                            check_key,
                            expected,
                            ip_version,
                            p.request_data,
                        )
                        if properties not in unique_properties:
                            unique_properties.append(properties)
                            task = Task(
                                domain=domain,
                                url=url,
                                ip_version=ip_version,
                                method=p.method,
                                request_data=p.request_data,
                                check=check_key,
                                expected=expected,
                                frequency=frequency,
                                recheck_delay=recheck_delay,
                                retry_before_notification=retry_before_notification,
                                already_retried=False,
                            )
                            logger.debug("Adding a new task in the db: %s", task)
                            tasks.append(task)

    db.add_all(tasks)
    db.commit()

    # Delete vanished tasks
    if max_task_id:
        vanished_tasks = (
            db.query(Task)
            .filter(Task.id <= max_task_id, Task.id.not_in(seen_tasks))
            .delete()
        )
        db.commit()
        logger.info(
            "%i tasks has been removed since not in config file anymore", vanished_tasks
        )
        return {"added": len(tasks), "vanished": vanished_tasks}

    return {"added": len(tasks), "vanished": 0}


async def get_severity_counts(db: Session) -> dict:
    """Get the severities (ok, warning, critical…) and their count"""
    query = db.query(Task.severity, func.count(Task.id).label("count")).group_by(  # pylint: disable-msg=not-callable
        Task.severity
    )

    # Execute the query and fetch the results
    task_counts_by_severity = query.all()

    counts_dict = dict(task_counts_by_severity)  # type: ignore[var-annotated,arg-type]
    for key in ("ok", "warning", "critical", "unknown"):
        counts_dict.setdefault(key, 0)
    return counts_dict


async def reschedule_all(db: Session):
    """Reschedule checks of all non OK tasks ASAP"""
    db.query(Task).filter(Task.severity.in_(["warning", "critical", "unknown"])).update(
        {Task.next_run: datetime.now() - timedelta(days=1)}
    )
    db.commit()


async def remove_old_results(db: Session, max_results: int):
    tasks = db.query(Task).all()
    deleted = 0
    for task in tasks:
        # Get the id of the oldest result to keep
        subquery = (
            db.query(Result.id)
            .filter(Result.task_id == task.id)
            .order_by(desc(Result.id))
            .limit(max_results)
            .subquery()
        )
        min_id = db.query(func.min(subquery.c.id)).scalar()  #  pylint: disable-msg=not-callable

        # Delete all the results older than min_id
        if min_id:
            deleted += (
                db.query(Result)
                .where(Result.id < min_id, Result.task_id == task.id)
                .delete()
            )
            db.commit()

    return deleted


async def release_old_locks(db: Session, max_lock_seconds: int):
    """Remove outdated locks on tasks"""
    max_acceptable_time = datetime.now() - timedelta(seconds=max_lock_seconds)

    # Release the locks on jobs that have been selected_at for more than max_lock_time
    updated = (
        db.query(Task)
        .filter(Task.selected_at < max_acceptable_time)
        .update({Task.selected_at: None, Task.selected_by: None})
    )
    db.commit()
    return updated


async def get_recent_agents_count(db: Session, minutes: int):
    """Get agents seen less than <minutes> ago"""
    max_time = datetime.now() - timedelta(minutes=minutes)

    agents = db.query(Result.agent_id).filter(Result.submitted_at > max_time).distinct()
    return agents.count()



import logging


logging.getLogger("passlib").setLevel(logging.ERROR)


LOG_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]

# Print level before message
logging.basicConfig(format="%(levelname)-9s %(message)s")

# XXX We probably want different loggers for client and server.
logger = logging.getLogger(__name__)


# XXX Does not work ?
def set_log_level(log_level):
    level = getattr(logging, log_level.upper(), None)
    if not isinstance(level, int):
        raise ValueError(f"Invalid log level: {log_level}")
    logger.setLevel(level=level)
    logger.info("Log level set to %s", log_level)



"""Add IP version to checks

Revision ID: 64f73a79b7d8
Revises: a1e98cf72a5c
Create Date: 2024-12-02 14:12:40.558033

"""
from typing import Sequence, Union

from alembic import op
from sqlalchemy.dialects.postgresql import ENUM
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "64f73a79b7d8"
down_revision: Union[str, None] = "a1e98cf72a5c"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    enum = ENUM("4", "6", name="ip_version_enum", create_type=False)
    enum.create(op.get_bind(), checkfirst=True)
    with op.batch_alter_table("tasks", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("ip_version", enum, server_default="4", nullable=False)
        )


def downgrade() -> None:
    with op.batch_alter_table("tasks", schema=None) as batch_op:
        batch_op.drop_column("ip_version")
        ENUM(name="ip_version_enum").drop(op.get_bind(), checkfirst=True)



from fastapi import Depends, HTTPException, Request
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
from fastapi_login import LoginManager

auth_scheme = HTTPBearer()


def get_db(request: Request):
    db = request.app.state.SessionLocal()
    try:
        yield db
    finally:
        db.close()


def get_config(request: Request):
    return request.app.state.config


async def get_manager(request: Request) -> LoginManager:
    if request.app.state.config.general.unauthenticated_access is not None:
        return await request.app.state.manager.optional(request)

    return await request.app.state.manager(request)


async def verify_token(
    request: Request, token: HTTPAuthorizationCredentials = Depends(auth_scheme)
):
    """Verify agent token"""
    if token.credentials not in request.app.state.config.service.secrets:
        raise HTTPException(status_code=401, detail="Unauthorized")
    return token


async def find_ldap_user(config, ldap, user: str) -> str | None:
    """Do a LDAP search for user and return its dn"""
    import ldap.filter as ldap_filter
    from ldapurl import LDAP_SCOPE_SUBTREE

    result = ldap.search_s(
        config.general.ldap.user_tree,
        LDAP_SCOPE_SUBTREE,
        filterstr=ldap_filter.filter_format(
            f"(&(%s=%s){config.general.ldap.user_filter})",
            [
                config.general.ldap.user_attr,
                user,
            ],
        ),
        attrlist=[config.general.ldap.user_attr],
    )

    # If there is a result, there should, logically, be only one entry
    if len(result) > 0:
        return result[0][0]

    return None



import os

from click.testing import CliRunner
from argos.commands import (
    add,
    verify_password,
    change_password,
    show,
    disable,
    enable,
    delete,
)


os.environ["ARGOS_APP_ENV"] = "test"
os.environ["ARGOS_YAML_FILE"] = "tests/config.yaml"


def test_add_user():
    runner = CliRunner()
    result = runner.invoke(add, ["--name", "foo"], input="bar\nbar\n")
    assert result.exit_code == 0
    assert result.output == "Password: \nRepeat for confirmation: \nUser foo added.\n"
    result = runner.invoke(add, ["--name", "foo"], input="bar\nbar\n")
    assert result.exit_code == 1
    assert (
        result.output
        == "Password: \nRepeat for confirmation: \nUser foo already exists.\n"
    )
    result = runner.invoke(add, ["--name", "baz", "--password", "qux"])
    assert result.exit_code == 0
    assert result.output == "User baz added.\n"


def test_user_password():
    runner = CliRunner()
    result = runner.invoke(verify_password, ["--name", "foo"], input="bar\n")
    assert result.exit_code == 0
    assert result.output == "Password: \nThe provided password is correct.\n"
    result = runner.invoke(verify_password, ["--name", "foo", "--password", "bar"])
    assert result.exit_code == 0
    assert result.output == "The provided password is correct.\n"
    result = runner.invoke(verify_password, ["--name", "quux", "--password", "corge"])
    assert result.exit_code == 1
    assert result.output == "User quux does not exist.\n"
    result = runner.invoke(verify_password, ["--name", "foo", "--password", "grault"])
    assert result.exit_code == 2
    assert result.output == "Wrong password!\n"


def test_change_password():
    runner = CliRunner()
    result = runner.invoke(verify_password, ["--name", "foo", "--password", "grault"])
    assert result.exit_code == 2
    assert result.output == "Wrong password!\n"
    result = runner.invoke(change_password, ["--name", "foo"], input="grault\ngrault\n")
    assert result.exit_code == 0
    assert (
        result.output
        == "Password: \nRepeat for confirmation: \nPassword of user foo changed.\n"
    )
    result = runner.invoke(verify_password, ["--name", "foo", "--password", "grault"])
    assert result.exit_code == 0
    assert result.output == "The provided password is correct.\n"
    result = runner.invoke(change_password, ["--name", "foo", "--password", "bar"])
    assert result.exit_code == 0
    assert result.output == "Password of user foo changed.\n"
    result = runner.invoke(verify_password, ["--name", "foo", "--password", "bar"])
    assert result.exit_code == 0
    assert result.output == "The provided password is correct.\n"
    result = runner.invoke(verify_password, ["--name", "quux", "--password", "bar"])
    assert result.exit_code == 1
    assert result.output == "User quux does not exist.\n"


def test_show():
    runner = CliRunner()
    result = runner.invoke(show)
    assert result.exit_code == 0
    assert (
        result.output
        == "✅ means that the user is enabled.\n❌ means that the user is disabled.\n"
        "✅ baz, last login: None\n✅ foo, last login: None\n"
    )


def test_disable():
    runner = CliRunner()
    result = runner.invoke(disable, ["--name", "quux"])
    assert result.exit_code == 1
    assert result.output == "User quux does not exist.\n"
    result = runner.invoke(disable, ["--name", "foo"])
    assert result.exit_code == 0
    assert result.output == "User foo disabled.\n"
    result = runner.invoke(disable, ["--name", "foo"])
    assert result.exit_code == 2
    assert result.output == "User foo is already disabled.\n"
    result = runner.invoke(show)
    assert result.exit_code == 0
    assert (
        result.output
        == "✅ means that the user is enabled.\n❌ means that the user is disabled.\n"
        "✅ baz, last login: None\n❌ foo, last login: None\n"
    )


def test_enable():
    runner = CliRunner()
    result = runner.invoke(enable, ["--name", "quux"])
    assert result.exit_code == 1
    assert result.output == "User quux does not exist.\n"
    result = runner.invoke(enable, ["--name", "foo"])
    assert result.exit_code == 0
    assert result.output == "User foo enabled.\n"
    result = runner.invoke(enable, ["--name", "foo"])
    assert result.exit_code == 2
    assert result.output == "User foo is already enabled.\n"
    result = runner.invoke(show)
    assert result.exit_code == 0
    assert (
        result.output
        == "✅ means that the user is enabled.\n❌ means that the user is disabled.\n"
        "✅ baz, last login: None\n✅ foo, last login: None\n"
    )


def test_delete():
    runner = CliRunner()
    result = runner.invoke(delete, ["--name", "quux"])
    assert result.exit_code == 1
    assert result.output == "User quux does not exist.\n"
    result = runner.invoke(delete, ["--name", "foo"])
    assert result.exit_code == 0
    assert result.output == "User foo deleted.\n"
    result = runner.invoke(delete, ["--name", "foo"])
    assert result.exit_code == 1
    assert result.output == "User foo does not exist.\n"
    result = runner.invoke(show)
    assert result.exit_code == 0
    assert (
        result.output
        == "✅ means that the user is enabled.\n❌ means that the user is disabled.\n"
        "✅ baz, last login: None\n"
    )
    result = runner.invoke(delete, ["--name", "baz"])
    assert result.exit_code == 0
    assert result.output == "User baz deleted.\n"
    result = runner.invoke(show)
    assert result.exit_code == 1
    assert result.output == "There is no users in database.\n"



from argos.checks.base import (  # NOQA
    BaseCheck,
    CheckNotFound,
    get_registered_check,
    get_registered_checks,
)
from argos.checks.checks import (  # NOQA
    HTTPBodyContains,
    HTTPStatus,
    SSLCertificateExpiration,
)



"""Define the available checks"""

import json
import re
from datetime import datetime

from httpx import Response
from jsonpointer import resolve_pointer, JsonPointerException

from argos.checks.base import (
    BaseCheck,
    ExpectedIntValue,
    ExpectedStringValue,
    Severity,
    Status,
)


class HTTPStatus(BaseCheck):
    """Checks that the HTTP status code is the expected one."""

    config = "status-is"
    expected_cls = ExpectedIntValue

    async def run(self, response: Response) -> dict:
        return self.response(
            status=response.status_code == self.expected,
            expected=self.expected,
            retrieved=response.status_code,
        )


class HTTPStatusIn(BaseCheck):
    """Checks that the HTTP status code is in the list of expected values."""

    config = "status-in"
    expected_cls = ExpectedStringValue

    async def run(self, response: Response) -> dict:
        return self.response(
            status=response.status_code in json.loads(self.expected),
            expected=self.expected,
            retrieved=response.status_code,
        )


class HTTPToHTTPS(BaseCheck):
    """Checks that the HTTP to HTTPS redirection status code is the expected one."""

    config = "http-to-https"
    expected_cls = ExpectedStringValue

    async def run(self, response: Response) -> dict:
        expected_dict = json.loads(self.expected)
        expected = range(300, 400)
        if "range" in expected_dict:
            expected = range(expected_dict["range"][0], expected_dict["range"][1])
        if "value" in expected_dict:
            expected = range(expected_dict["value"], expected_dict["value"] + 1)
        if "list" in expected_dict:
            expected = expected_dict["list"]

        return self.response(
            status=response.status_code in expected,
            expected=self.expected,
            retrieved=response.status_code,
        )


class HTTPHeadersContain(BaseCheck):
    """Checks that response headers contains the expected headers
    (without checking their values)"""

    config = "headers-contain"
    expected_cls = ExpectedStringValue

    async def run(self, response: Response) -> dict:
        status = True
        for header in json.loads(self.expected):
            if header not in response.headers:
                status = False
                break

        return self.response(
            status=status,
            expected=self.expected,
            retrieved=json.dumps(list(dict(response.headers).keys())),
        )


class HTTPHeadersHave(BaseCheck):
    """Checks that response headers contains the expected headers and values"""

    config = "headers-have"
    expected_cls = ExpectedStringValue

    async def run(self, response: Response) -> dict:
        status = True
        for header, value in json.loads(self.expected).items():
            if header not in response.headers:
                status = False
                break
            if response.headers[header] != value:
                status = False
                break

        return self.response(
            status=status,
            expected=self.expected,
            retrieved=json.dumps(dict(response.headers)),
        )


class HTTPHeadersLike(BaseCheck):
    """Checks that response headers contains the expected headers and that the values
    matches the provided regexes"""

    config = "headers-like"
    expected_cls = ExpectedStringValue

    async def run(self, response: Response) -> dict:
        status = True
        for header, value in json.loads(self.expected).items():
            if header not in response.headers:
                status = False
                break
            if not re.search(rf"{value}", response.headers[header]):
                status = False
                break

        return self.response(
            status=status,
            expected=self.expected,
            retrieved=json.dumps(dict(response.headers)),
        )


class HTTPBodyContains(BaseCheck):
    """Checks that the HTTP body contains the expected string."""

    config = "body-contains"
    expected_cls = ExpectedStringValue

    async def run(self, response: Response) -> dict:
        return self.response(status=self.expected in response.text)


class HTTPBodyLike(BaseCheck):
    """Checks that the HTTP body matches the provided regex."""

    config = "body-like"
    expected_cls = ExpectedStringValue

    async def run(self, response: Response) -> dict:
        if re.search(rf"{self.expected}", response.text):
            return self.response(status=True)

        return self.response(status=False)


class HTTPJsonContains(BaseCheck):
    """Checks that JSON response contains the expected structure
    (without checking the value)"""

    config = "json-contains"
    expected_cls = ExpectedStringValue

    async def run(self, response: Response) -> dict:
        obj = response.json()

        status = True
        for pointer in json.loads(self.expected):
            try:
                resolve_pointer(obj, pointer)
            except JsonPointerException:
                status = False
                break

        return self.response(
            status=status,
            expected=self.expected,
            retrieved=json.dumps(obj),
        )


class HTTPJsonHas(BaseCheck):
    """Checks that JSON response contains the expected structure and values"""

    config = "json-has"
    expected_cls = ExpectedStringValue

    async def run(self, response: Response) -> dict:
        obj = response.json()

        status = True
        for pointer, exp_value in json.loads(self.expected).items():
            try:
                value = resolve_pointer(obj, pointer)
                if value != exp_value:
                    status = False
                    break
            except JsonPointerException:
                status = False
                break

        return self.response(
            status=status,
            expected=self.expected,
            retrieved=json.dumps(obj),
        )


class HTTPJsonLike(BaseCheck):
    """Checks that JSON response contains the expected structure and that the values
    matches the provided regexes"""

    config = "json-like"
    expected_cls = ExpectedStringValue

    async def run(self, response: Response) -> dict:
        obj = response.json()

        status = True
        for pointer, exp_value in json.loads(self.expected).items():
            try:
                value = resolve_pointer(obj, pointer)
                if not re.search(rf"{exp_value:}", value):
                    status = False
                    break
            except JsonPointerException:
                status = False
                break

        return self.response(
            status=status,
            expected=self.expected,
            retrieved=json.dumps(obj),
        )


class HTTPJsonIs(BaseCheck):
    """Checks that JSON response is the exact expected JSON object"""

    config = "json-is"
    expected_cls = ExpectedStringValue

    async def run(self, response: Response) -> dict:
        obj = response.json()

        status = response.json() == json.loads(self.expected)

        return self.response(
            status=status,
            expected=self.expected,
            retrieved=json.dumps(obj),
        )


class SSLCertificateExpiration(BaseCheck):
    """Checks that the SSL certificate will not expire soon."""

    config = "ssl-certificate-expiration"
    expected_cls = ExpectedStringValue

    async def run(self, response: Response) -> dict:
        """Returns the number of days in which the certificate will expire."""
        network_stream = response.extensions["network_stream"]
        ssl_obj = network_stream.get_extra_info("ssl_object")
        cert = ssl_obj.getpeercert()

        not_after = datetime.strptime(cert.get("notAfter"), "%b %d %H:%M:%S %Y %Z")
        expires_in = (not_after - datetime.now()).days

        return self.response(status=Status.ON_CHECK, expires_in=expires_in)

    @classmethod
    async def finalize(cls, config, result, **context):
        if result.status == Status.ERROR:
            return result.status, Severity.UNKNOWN
        if result.status != Status.ON_CHECK:
            return result.status, Severity.WARNING

        if "expires_in" in context:
            thresholds = config.ssl.thresholds
            thresholds.sort()
            for days, severity in thresholds:
                if context["expires_in"] < days:
                    return Status.FAILURE, severity
            return Status.SUCCESS, Severity.OK

        raise ValueError(
            "The SSLCertificateExpiration check didn't provide an 'expires_in' "
            "context variable."
        )

    @classmethod
    def get_description(cls, config):
        thresholds = config.ssl.thresholds
        thresholds.sort()
        string = SSLCertificateExpiration.__doc__ + "\n"
        for days, severity in thresholds:
            string += f"- {severity} if expiration in less than {days} days\n"
        return string



"""Adding ConfigCache model

Revision ID: 1a3497f9f71b
Revises: e99bc35702c9
Create Date: 2024-03-13 15:28:09.185377

"""
from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision: str = "1a3497f9f71b"
down_revision: Union[str, None] = "e99bc35702c9"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.create_table(
        "config_cache",
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("val", sa.String(), nullable=False),
        sa.Column("updated_at", sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint("name"),
    )


def downgrade() -> None:
    op.drop_table("config_cache")



"""Various base classes for checks"""

from dataclasses import dataclass
from typing import Type

from pydantic import BaseModel

from argos.schemas.models import Task


class Status:
    """Possible statuses of the checks"""

    ON_CHECK = "on-check"
    SUCCESS = "success"
    FAILURE = "failure"
    ERROR = "error"


class Severity:
    """Possible statuses of the checks’ results"""

    OK = "ok"
    WARNING = "warning"
    CRITICAL = "critical"
    UNKNOWN = "unknown"


# XXX We could name this Result, but is it could overlap with schemas.Result.
# Need to better define the naming around this.
# Status can be "Success" / "Failure" / "Error" or "On Check"
@dataclass
class Response:
    status: str
    context: dict

    @classmethod
    def new(cls, status, **kwargs):
        """Normalize results of checks."""
        if isinstance(status, bool):
            status = Status.SUCCESS if status else Status.FAILURE

        return cls(status=status, context=kwargs)


class BaseExpectedValue(BaseModel):
    expected: str

    def get_converted(self):
        return self.expected


class ExpectedIntValue(BaseExpectedValue):
    def get_converted(self):
        return int(self.expected)


class ExpectedStringValue(BaseExpectedValue):
    pass


class CheckNotFound(Exception):
    pass


class InvalidResponse(Exception):
    def __str__(self):
        return "The provided response is missing a 'status' key."


class BaseCheck:
    config: str
    expected_cls: None | Type[BaseExpectedValue] = None

    _registry = []  # type: ignore[var-annotated]

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._registry.append(cls)

    @classmethod
    def get_registered_checks(cls):
        """Return existing checks"""
        return {c.config: c for c in cls._registry}

    @classmethod
    def get_registered_check(cls, name):
        """Get a check from its name"""
        check = cls.get_registered_checks().get(name)
        if not check:
            raise CheckNotFound(name)
        return check

    def __init__(self, task: Task):
        self.task = task

    @property
    def expected(self):
        """Convert the task’s class to simpler class"""
        if self.expected_cls is not None:
            return self.expected_cls(expected=self.task.expected).get_converted()
        return None

    def response(self, **kwargs):
        """Ensure that the response has a status and return a Response"""
        if "status" not in kwargs:
            raise InvalidResponse(kwargs)
        status = kwargs.pop("status")
        return Response.new(status, **kwargs)

    @classmethod
    async def finalize(cls, config, result, **context):
        """By default, the finalize considers that :

        - All FAILUREs should be reported as CRITICAL
        - All SUCCESS should be reported as OK
        - All ERRORS should be reported as UNKNOWN.

        This behaviour can be changed in each check, by defining the `finalize` method.
        XXX Allow this to be tweaked by the config.
        """
        if result.status == Status.SUCCESS:
            return result.status, Severity.OK
        if result.status == Status.ERROR:
            return result.status, Severity.UNKNOWN
        if result.status == Status.FAILURE:
            return result.status, Severity.CRITICAL
        if result.status == Status.ON_CHECK:
            msg = (
                "Status is 'on-check', but the Check class "
                "didn't provide a finalize() method."
            )
            raise ValueError(msg)

    @classmethod
    def get_description(cls, config):
        return cls.__doc__ or ""


def get_registered_check(name):
    return BaseCheck.get_registered_check(name)


def get_registered_checks():
    return BaseCheck.get_registered_checks()



"""Pydantic schemas for server"""
from pathlib import Path

import yaml
from yamlinclude import YamlIncludeConstructor

from argos.schemas.config import Config


def read_yaml_config(filename: str) -> Config:
    parsed = _load_yaml(filename)
    return Config(**parsed)


def _load_yaml(filename: str):
    base_dir = Path(filename).resolve().parent
    YamlIncludeConstructor.add_to_loader_class(
        loader_class=yaml.FullLoader, base_dir=str(base_dir)
    )

    with open(filename, "r", encoding="utf-8") as stream:
        return yaml.load(stream, Loader=yaml.FullLoader)



"""Add task index

Revision ID: 8b58ced14d6e
Revises: 64f73a79b7d8
Create Date: 2024-12-03 16:41:44.842213

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "8b58ced14d6e"
down_revision: Union[str, None] = "64f73a79b7d8"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    with op.batch_alter_table("tasks", schema=None) as batch_op:
        batch_op.add_column(sa.Column("task_group", sa.String(), nullable=True))
    with op.batch_alter_table("tasks", schema=None) as batch_op:
        batch_op.execute(
            "UPDATE tasks SET task_group = method || '-' || ip_version || '-' || url"
        )
        batch_op.alter_column("task_group", nullable=False)
        batch_op.create_index("similar_tasks", ["task_group"], unique=False)


def downgrade() -> None:
    with op.batch_alter_table("tasks", schema=None) as batch_op:
        batch_op.drop_index("similar_tasks")
        batch_op.drop_column("task_group")



"""Argos agent

Fetchs the tasks from the server, execute it and send the result to the server
"""
import asyncio
import json
import logging
import socket
from hashlib import md5
from time import sleep
from typing import List

import httpx
from tenacity import retry, wait_random  # type: ignore

from argos import VERSION
from argos.checks import get_registered_check
from argos.logging import logger
from argos.schemas import AgentResult, SerializableException, Task


def log_failure(retry_state):
    """Log failures, with a different log level depending on the number of attempts."""
    if retry_state.attempt_number < 1:
        loglevel = logging.INFO
    else:
        loglevel = logging.WARNING
    logger.log(
        loglevel,
        "Retrying: attempt %s ended with: %s %s",
        retry_state.attempt_number,
        retry_state.outcome,
        retry_state.outcome.exception(),
    )


class ArgosAgent:  # pylint: disable-msg=too-many-instance-attributes
    """The Argos agent is responsible for running the checks and reporting the results."""

    def __init__(self, server: str, auth: str, max_tasks: int, wait_time: int):
        self.server = server
        self.max_tasks = max_tasks
        self.wait_time = wait_time
        self.auth = auth
        self._http_client: httpx.AsyncClient | None = None
        self._http_client_v4: httpx.AsyncClient | None = None
        self._http_client_v6: httpx.AsyncClient | None = None
        self._res_cache: dict[str, httpx.Response] = {}

        self.agent_id = socket.gethostname()

    @retry(after=log_failure, wait=wait_random(min=1, max=2))
    async def run(self):
        auth_header = {
            "Authorization": f"Bearer {self.auth}",
            "User-Agent": f"Argos Panoptes agent {VERSION}",
        }
        self._http_client = httpx.AsyncClient(headers=auth_header)

        ua_header = {
            "User-Agent": f"Argos Panoptes {VERSION} "
            "(about: https://argos-monitoring.framasoft.org/)",
        }
        self._http_client_v4 = httpx.AsyncClient(
            headers=ua_header,
            transport=httpx.AsyncHTTPTransport(local_address="0.0.0.0"),
        )
        self._http_client_v6 = httpx.AsyncClient(
            headers=ua_header, transport=httpx.AsyncHTTPTransport(local_address="::")
        )

        logger.info("Running agent against %s", self.server)
        async with self._http_client:
            while "forever":
                retry_now = await self._get_and_complete_tasks()
                if not retry_now:
                    logger.info("Waiting %i seconds before next retry", self.wait_time)
                    await asyncio.sleep(self.wait_time)

    async def _do_request(self, group: str, details: dict):
        headers = {}
        if details["request_data"] is not None:
            request_data = json.loads(details["request_data"])
            if request_data["headers"] is not None:
                headers = request_data["headers"]

        if details["ip_version"] == "4":
            http_client = self._http_client_v4
        else:
            http_client = self._http_client_v6
        try:
            if details["request_data"] is None or request_data["data"] is None:
                response = await http_client.request(  # type: ignore[union-attr]
                    method=details["method"],
                    url=details["url"],
                    headers=headers,
                    timeout=60,
                )
            elif request_data["json"]:
                response = await http_client.request(  # type: ignore[union-attr]
                    method=details["method"],
                    url=details["url"],
                    headers=headers,
                    json=request_data["data"],
                    timeout=60,
                )
            else:
                response = await http_client.request(  # type: ignore[union-attr]
                    method=details["method"],
                    url=details["url"],
                    headers=headers,
                    data=request_data["data"],
                    timeout=60,
                )
        except httpx.ReadError:
            sleep(1)
            if details["request_data"] is None or request_data["data"] is None:
                response = await http_client.request(  # type: ignore[union-attr]
                    method=details["method"], url=details["url"], timeout=60
                )
            elif request_data["json"]:
                response = await http_client.request(  # type: ignore[union-attr]
                    method=details["method"],
                    url=details["url"],
                    json=request_data["data"],
                    timeout=60,
                )
            else:
                response = await http_client.request(  # type: ignore[union-attr]
                    method=details["method"],
                    url=details["url"],
                    data=request_data["data"],
                    timeout=60,
                )

        self._res_cache[group] = response

    async def _complete_task(self, _task: dict) -> AgentResult:
        try:
            task = Task(**_task)

            check_class = get_registered_check(task.check)
            check = check_class(task)
            result = await check.run(self._res_cache[task.task_group])
            status = result.status
            context = result.context

        except Exception as err:  # pylint: disable=broad-except
            status = "error"
            context = SerializableException.from_exception(err)
            msg = f"An exception occured when running {_task}. {err.__class__.__name__} : {err}"
            logger.error(msg)
        return AgentResult(task_id=task.id, status=status, context=context)

    async def _get_and_complete_tasks(self):
        # Fetch the list of tasks
        response = await self._http_client.get(
            f"{self.server}/api/tasks",
            params={"limit": self.max_tasks, "agent_id": self.agent_id},
        )

        if response.status_code == httpx.codes.OK:
            data = response.json()
            logger.info("Received %i tasks from the server", len(data))

            req_groups = {}
            _tasks = []
            for _task in data:
                task = Task(**_task)

                url = task.url
                group = task.task_group

                if task.check == "http-to-https":
                    data = task.request_data
                    if data is None:
                        data = ""
                    url = str(httpx.URL(task.url).copy_with(scheme="http"))
                    group = (
                        f"{task.method}-{task.ip_version}-{url}-"
                        f"{md5(data.encode()).hexdigest()}"
                    )
                    _task["task_group"] = group

                req_groups[group] = {
                    "url": url,
                    "ip_version": task.ip_version,
                    "method": task.method,
                    "request_data": task.request_data,
                }
                _tasks.append(_task)

            requests = []
            for group, details in req_groups.items():
                requests.append(self._do_request(group, details))

            if requests:
                await asyncio.gather(*requests)

            tasks = []
            for task in _tasks:
                tasks.append(self._complete_task(task))

            if tasks:
                results = await asyncio.gather(*tasks)
                await self._post_results(results)
                return True

            logger.info("Got no tasks from the server.")
            return False

        logger.error("Failed to fetch tasks: %s", response.read())
        return False

    async def _post_results(self, results: List[AgentResult]):
        data = [r.model_dump() for r in results]
        if self._http_client is not None:
            response = await self._http_client.post(
                f"{self.server}/api/results",
                params={"agent_id": self.agent_id},
                json=data,
            )

            if response.status_code == httpx.codes.CREATED:
                logger.info(
                    "Successfully posted results %s", json.dumps(response.json())
                )
            else:
                logger.error("Failed to post results: %s", response.read())
            return response

        logger.error("self._http_client is None")



"""Database models"""

from datetime import datetime, timedelta
from hashlib import md5
from typing import List, Literal

from sqlalchemy import (
    JSON,
    Enum,
    ForeignKey,
)
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship
from sqlalchemy.schema import Index

from argos.checks import BaseCheck, get_registered_check
from argos.schemas import WebsiteCheck
from argos.schemas.utils import IPVersion, Method


def compute_task_group(context) -> str:
    data = context.current_parameters["request_data"]
    if data is None:
        data = ""
    return (
        f"{context.current_parameters['method']}-"
        f"{context.current_parameters['ip_version']}-"
        f"{context.current_parameters['url']}-"
        f"{md5(data.encode()).hexdigest()}"
    )


class Base(DeclarativeBase):
    type_annotation_map = {List[WebsiteCheck]: JSON, dict: JSON}


class Task(Base):
    """
    There is one task per check.

    It contains all information needed to run the jobs on the agents.
    Agents will return information in the result table.
    """

    __tablename__ = "tasks"
    id: Mapped[int] = mapped_column(primary_key=True)

    # Info needed to run the task
    url: Mapped[str] = mapped_column()
    domain: Mapped[str] = mapped_column()
    ip_version: Mapped[IPVersion] = mapped_column(
        Enum("4", "6", name="ip_version_enum"),
    )
    check: Mapped[str] = mapped_column()
    expected: Mapped[str] = mapped_column()
    frequency: Mapped[float] = mapped_column()
    recheck_delay: Mapped[float] = mapped_column(nullable=True)
    already_retried: Mapped[bool] = mapped_column(insert_default=False)
    retry_before_notification: Mapped[int] = mapped_column(insert_default=0)
    contiguous_failures: Mapped[int] = mapped_column(insert_default=0)
    method: Mapped[Method] = mapped_column(
        Enum(
            "GET",
            "HEAD",
            "POST",
            "OPTIONS",
            "CONNECT",
            "TRACE",
            "PUT",
            "PATCH",
            "DELETE",
            name="method",
        ),
        insert_default="GET",
    )
    request_data: Mapped[str] = mapped_column(nullable=True)

    # Orchestration-related
    selected_by: Mapped[str] = mapped_column(nullable=True)
    selected_at: Mapped[datetime] = mapped_column(nullable=True)
    completed_at: Mapped[datetime] = mapped_column(nullable=True)
    next_run: Mapped[datetime] = mapped_column(nullable=True)
    task_group: Mapped[str] = mapped_column(insert_default=compute_task_group)

    severity: Mapped[Literal["ok", "warning", "critical", "unknown"]] = mapped_column(
        Enum("ok", "warning", "critical", "unknown", name="severity"),
        insert_default="unknown",
    )
    last_severity_update: Mapped[datetime] = mapped_column(nullable=True)

    results: Mapped[List["Result"]] = relationship(
        back_populates="task",
        cascade="all, delete",
        passive_deletes=True,
    )

    def __str__(self) -> str:
        return f"DB Task {self.url} (IPv{self.ip_version}) - {self.check} - {self.expected}"

    def get_check(self) -> BaseCheck:
        """Returns a check instance for this specific task"""
        return get_registered_check(self.check)

    def set_times_severity_and_deselect(self, severity, submitted_at):
        """Removes the lock on task, set its severity and set the time for the next run"""
        self.severity = severity
        self.last_severity_update = submitted_at
        self.selected_by = None
        self.selected_at = None

        now = datetime.now()
        self.completed_at = now
        if (
            self.recheck_delay is not None
            and severity != "ok"
            and not self.already_retried
        ):
            self.next_run = now + timedelta(minutes=self.recheck_delay)
            self.already_retried = True
        else:
            self.next_run = now + timedelta(minutes=self.frequency)
            self.already_retried = False

    @property
    def last_result(self):
        """Get last result of the task"""
        if not self.results:
            return None
        return max(self.results, key=lambda r: r.id)

    @property
    def status(self):
        """Get status of the task, i.e. the status of its last result"""
        if not self.last_result:
            return None
        return self.last_result.status


Index("similar_tasks", Task.task_group)


class Result(Base):
    """There are multiple results per task.

    The results store information returned by the agents.

    You can read `status` as  "Was the agent able to do the check?"
    while the `severity` depends on the return value of the check.
    """

    __tablename__ = "results"
    id: Mapped[int] = mapped_column(primary_key=True)
    task_id: Mapped[int] = mapped_column(ForeignKey("tasks.id", ondelete="CASCADE"))
    task: Mapped["Task"] = relationship(back_populates="results")
    agent_id: Mapped[str] = mapped_column(nullable=True)

    submitted_at: Mapped[datetime] = mapped_column()
    # XXX change "on-check" to something better.
    status: Mapped[Literal["success", "failure", "error", "on-check"]] = mapped_column(
        Enum("success", "failure", "error", "on-check", name="status")
    )
    severity: Mapped[Literal["ok", "warning", "critical", "unknown"]] = mapped_column(
        Enum("ok", "warning", "critical", "unknown", name="severity")
    )
    context: Mapped[dict] = mapped_column()

    def set_status(self, status, severity):
        self.severity = severity
        self.status = status

    def __str__(self):
        return f"DB Result {self.id} - {self.status} - {self.context}"


class ConfigCache(Base):
    """Database model containing information on the current state
    of the configuration.

    This is used to determine if tasks are to be updated.

    These settings are cached:
    - general_frequency: the content of general.frequency setting, in minutes
                         ex: 5
    - websites_hash:     the hash (sha256sum) of websites setting, to allow a quick
                         comparison without looping through all websites.
                         ex: 8b886e7db7b553fe99f6d5437f31745987e243c77b2109b84cf9a7f8bf7d75b1
    """

    __tablename__ = "config_cache"
    name: Mapped[str] = mapped_column(primary_key=True)
    val: Mapped[str] = mapped_column()
    updated_at: Mapped[datetime] = mapped_column()


class User(Base):
    """Database model for user authentication"""

    __tablename__ = "users"
    username: Mapped[str] = mapped_column(primary_key=True)
    password: Mapped[str] = mapped_column()
    disabled: Mapped[bool] = mapped_column()
    created_at: Mapped[datetime] = mapped_column(default=datetime.now())
    updated_at: Mapped[datetime] = mapped_column(nullable=True)
    last_login_at: Mapped[datetime] = mapped_column(nullable=True)

    def update_last_login_at(self):
        self.last_login_at = datetime.now()



import asyncio
import os
from functools import wraps
from pathlib import Path
from sys import exit as sysexit
from uuid import uuid4

import click
import uvicorn
from alembic import command
from alembic.config import Config

from argos import VERSION, logging
from argos.agent import ArgosAgent


async def get_db():
    from argos.server.main import connect_to_db, get_application, setup_database

    app = get_application()
    setup_database(app)
    return await connect_to_db(app)


def coroutine(f):
    """Decorator to enable async functions in click"""

    @wraps(f)
    def wrapper(*args, **kwargs):
        return asyncio.run(f(*args, **kwargs))

    return wrapper


def validate_config_access(ctx, param, value):
    for file in list(
        dict.fromkeys([value, "argos-config.yaml", "/etc/argos/config.yaml"])
    ):
        path = Path(file)

        if path.is_file() and os.access(path, os.R_OK):
            return file

    if value == "argos-config.yaml":
        raise click.BadParameter(
            f"the file {value} does not exists or is not reachable, "
            "nor does /etc/argos/config.yaml."
        )

    raise click.BadParameter(
        f"the file {value} does not exists or is not reachable, "
        "nor does argos-config.yaml or /etc/argos/config.yaml."
    )


@click.group()
def cli():
    pass


@cli.group()
def server():
    """Commands for managing server, server’s configuration and users"""


@server.group()
def user():
    """User management"""


@cli.command()
def version():
    """Prints Argos’ version and exits"""
    click.echo(VERSION)


@cli.command()
@click.argument("server_url", envvar="ARGOS_AGENT_SERVER_URL")
@click.argument("auth", envvar="ARGOS_AGENT_TOKEN")
@click.option(
    "--max-tasks",
    default=10,
    help="Number of concurrent tasks this agent can run",
)
@click.option(
    "--wait-time",
    default=10,
    help="Waiting time between two polls on the server (seconds)",
)
@click.option(
    "--log-level",
    default="INFO",
    type=click.Choice(logging.LOG_LEVELS, case_sensitive=False),
)
def agent(server_url, auth, max_tasks, wait_time, log_level):
    """Get and run tasks for the provided server. Will wait for new tasks.

    Usage: argos agent https://argos.example.org "auth-token-here"

    Alternatively, you can use the following environment variables to avoid passing
    arguments to the agent on the command line:

      \b
      ARGOS_AGENT_SERVER_URL=https://argos.example.org
      ARGOS_AGENT_TOKEN=auth-token-here
    """
    click.echo("Starting argos agent. Will retry forever.")
    from argos.logging import logger

    logger.setLevel(log_level)
    agent_ = ArgosAgent(server_url, auth, max_tasks, wait_time)
    asyncio.run(agent_.run())


@server.command()
@click.option("--host", default="127.0.0.1", help="Host to bind")
@click.option("--port", default=8000, type=int, help="Port to bind")
@click.option(
    "--config",
    default="argos-config.yaml",
    help="Path of the configuration file. "
    "If ARGOS_YAML_FILE environment variable is set, its value will be used instead. "
    "Default value: argos-config.yaml and /etc/argos/config.yaml as fallback.",
    envvar="ARGOS_YAML_FILE",
    callback=validate_config_access,
)
@click.option("--reload", is_flag=True, help="Enable hot reloading")
def start(host, port, config, reload):
    """Starts the server (use only for testing or development!)

    See https://argos-monitoring.framasoft.org/deployment/systemd.html#server
    for advices on how to start the server for production.
    """
    os.environ["ARGOS_YAML_FILE"] = config
    uvicorn.run("argos.server:app", host=host, port=port, reload=reload)


def validate_max_lock_seconds(ctx, param, value):
    if value <= 60:
        raise click.BadParameter("Should be strictly higher than 60")
    return value


def validate_max_results(ctx, param, value):
    if value <= 0:
        raise click.BadParameter("Should be a positive integer")
    return value


@server.command()
@click.option(
    "--max-results",
    default=100,
    help="Number of results per task to keep",
    callback=validate_max_results,
)
@click.option(
    "--max-lock-seconds",
    default=100,
    help=(
        "The number of seconds after which a lock is "
        "considered stale, must be higher than 60 "
        "(the checks have a timeout value of 60 seconds)"
    ),
    callback=validate_max_lock_seconds,
)
@click.option(
    "--config",
    default="argos-config.yaml",
    help="Path of the configuration file. "
    "If ARGOS_YAML_FILE environment variable is set, its value will be used instead. "
    "Default value: argos-config.yaml and /etc/argos/config.yaml as fallback.",
    envvar="ARGOS_YAML_FILE",
    callback=validate_config_access,
)
@coroutine
async def cleandb(max_results, max_lock_seconds, config):
    """Clean the database (to run routinely)

    \b
    - Removes old results from the database.
    - Removes locks from tasks that have been locked for too long.
    """
    # It’s mandatory to do it before the imports
    os.environ["ARGOS_YAML_FILE"] = config

    # The imports are made here otherwise the agent will need server configuration files.
    from argos.server import queries

    db = await get_db()
    removed = await queries.remove_old_results(db, max_results)
    updated = await queries.release_old_locks(db, max_lock_seconds)

    click.echo(f"{removed} results removed")
    click.echo(f"{updated} locks released")


@server.command()
@click.option(
    "--time-without-agent",
    default=5,
    help="Time without seeing an agent after which a warning will be issued, in minutes. "
    "Default is 5 minutes.",
    callback=validate_max_results,
)
@click.option(
    "--config",
    default="argos-config.yaml",
    help="Path of the configuration file. "
    "If ARGOS_YAML_FILE environment variable is set, its value will be used instead.",
    envvar="ARGOS_YAML_FILE",
    callback=validate_config_access,
)
@coroutine
async def watch_agents(time_without_agent, config):
    """Watch agents (to run routinely)

    Issues a warning if no agent has been seen by the server for a given time.
    """
    # It’s mandatory to do it before the imports
    os.environ["ARGOS_YAML_FILE"] = config

    # The imports are made here otherwise the agent will need server configuration files.
    from argos.server import queries

    db = await get_db()
    agents = await queries.get_recent_agents_count(db, time_without_agent)
    if agents == 0:
        click.echo(f"No agent has been seen in the last {time_without_agent} minutes.")
        sysexit(1)


@server.command(short_help="Load or reload tasks’ configuration")
@click.option(
    "--config",
    default="argos-config.yaml",
    help="Path of the configuration file. "
    "If ARGOS_YAML_FILE environment variable is set, its value will be used instead. "
    "Default value: argos-config.yaml and /etc/argos/config.yaml as fallback.",
    envvar="ARGOS_YAML_FILE",
    callback=validate_config_access,
)
@coroutine
async def reload_config(config):
    """Read tasks’ configuration and add/delete tasks in database if needed"""
    # It’s mandatory to do it before the imports
    os.environ["ARGOS_YAML_FILE"] = config

    # The imports are made here otherwise the agent will need server configuration files.
    from argos.server import queries
    from argos.server.main import read_config

    _config = read_config(config)

    db = await get_db()
    changed = await queries.update_from_config(db, _config)

    click.echo(f"{changed['added']} tasks added")
    click.echo(f"{changed['vanished']} tasks deleted")


@server.command()
@click.option(
    "--config",
    default="argos-config.yaml",
    help="Path of the configuration file. "
    "If ARGOS_YAML_FILE environment variable is set, its value will be used instead. "
    "Default value: argos-config.yaml and /etc/argos/config.yaml as fallback.",
    envvar="ARGOS_YAML_FILE",
    callback=validate_config_access,
)
@coroutine
async def migrate(config):
    """Run database migrations"""
    # It’s mandatory to do it before the imports
    os.environ["ARGOS_YAML_FILE"] = config

    # The imports are made here otherwise the agent will need server configuration files.
    from argos.server.settings import read_yaml_config

    settings = read_yaml_config(config)

    current_dir = Path(__file__).resolve().parent
    alembic_cfg = Config(current_dir / "server" / "migrations" / "alembic.ini")
    alembic_cfg.set_main_option("sqlalchemy.url", str(settings.general.db.url))
    command.upgrade(alembic_cfg, "head")


@user.command()
@click.option(
    "--config",
    default="argos-config.yaml",
    help="Path of the configuration file. "
    "If ARGOS_YAML_FILE environment variable is set, its value will be used instead.",
    envvar="ARGOS_YAML_FILE",
    callback=validate_config_access,
)
@click.option("--name", prompt=True, help="Name of the user to create.")
@click.password_option()
@coroutine
async def add(config, name, password):
    """Add new user"""
    # It’s mandatory to do it before the imports
    os.environ["ARGOS_YAML_FILE"] = config

    # The imports are made here otherwise the agent will need server configuration files.
    from passlib.context import CryptContext

    from argos.server import queries

    db = await get_db()
    _user = await queries.get_user(db, name)
    if _user is not None:
        click.echo(f"User {name} already exists.")
        sysexit(1)

    pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")
    await queries.add_user(db, name, pwd_context.hash(password))
    click.echo(f"User {name} added.")


@user.command()
@click.option(
    "--config",
    default="argos-config.yaml",
    help="Path of the configuration file. "
    "If ARGOS_YAML_FILE environment variable is set, its value will be used instead.",
    envvar="ARGOS_YAML_FILE",
    callback=validate_config_access,
)
@click.option(
    "--name", prompt=True, help="Name of the user you want to change the password."
)
@click.password_option()
@coroutine
async def change_password(config, name, password):
    """Change user’s password"""
    # It’s mandatory to do it before the imports
    os.environ["ARGOS_YAML_FILE"] = config

    # The imports are made here otherwise the agent will need server configuration files.
    from passlib.context import CryptContext

    from argos.server import queries

    db = await get_db()
    _user = await queries.get_user(db, name)
    if _user is None:
        click.echo(f"User {name} does not exist.")
        sysexit(1)

    pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")
    _user.password = pwd_context.hash(password)
    db.commit()
    click.echo(f"Password of user {name} changed.")


@user.command()
@click.option(
    "--config",
    default="argos-config.yaml",
    help="Path of the configuration file. "
    "If ARGOS_YAML_FILE environment variable is set, its value will be used instead.",
    envvar="ARGOS_YAML_FILE",
    callback=validate_config_access,
)
@click.option(
    "--name", required=True, help="Name of the user you want to test the password for."
)
@click.option("--password", prompt=True, hide_input=True)
@coroutine
async def verify_password(config, name, password):
    """Test user’s password"""
    # It’s mandatory to do it before the imports
    os.environ["ARGOS_YAML_FILE"] = config

    # The imports are made here otherwise the agent will need server configuration files.
    from passlib.context import CryptContext

    from argos.server import queries

    db = await get_db()
    _user = await queries.get_user(db, name)
    if _user is None:
        click.echo(f"User {name} does not exist.")
        sysexit(1)

    pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")
    if not pwd_context.verify(password, _user.password):
        click.echo("Wrong password!")
        sysexit(2)

    click.echo("The provided password is correct.")


@user.command()
@click.option(
    "--config",
    default="argos-config.yaml",
    help="Path of the configuration file. "
    "If ARGOS_YAML_FILE environment variable is set, its value will be used instead.",
    envvar="ARGOS_YAML_FILE",
    callback=validate_config_access,
)
@click.option("--name", required=True, help="Name of the user to disable.")
@coroutine
async def disable(config, name):
    """Disable user"""
    # It’s mandatory to do it before the imports
    os.environ["ARGOS_YAML_FILE"] = config

    # The imports are made here otherwise the agent will need server configuration files.
    from argos.server import queries

    db = await get_db()
    _user = await queries.get_user(db, name)
    if _user is None:
        click.echo(f"User {name} does not exist.")
        sysexit(1)
    if _user.disabled:
        click.echo(f"User {name} is already disabled.")
        sysexit(2)

    _user.disabled = True
    db.commit()

    click.echo(f"User {name} disabled.")


@user.command()
@click.option(
    "--config",
    default="argos-config.yaml",
    help="Path of the configuration file. "
    "If ARGOS_YAML_FILE environment variable is set, its value will be used instead.",
    envvar="ARGOS_YAML_FILE",
    callback=validate_config_access,
)
@click.option("--name", required=True, help="Name of the user to reenable")
@coroutine
async def enable(config, name):
    """Enable user"""
    # It’s mandatory to do it before the imports
    os.environ["ARGOS_YAML_FILE"] = config

    # The imports are made here otherwise the agent will need server configuration files.
    from argos.server import queries

    db = await get_db()
    _user = await queries.get_user(db, name)
    if _user is None:
        click.echo(f"User {name} does not exist.")
        sysexit(1)
    if not _user.disabled:
        click.echo(f"User {name} is already enabled.")
        sysexit(2)

    _user.disabled = False
    db.commit()

    click.echo(f"User {name} enabled.")


@user.command()
@click.option(
    "--config",
    default="argos-config.yaml",
    help="Path of the configuration file. "
    "If ARGOS_YAML_FILE environment variable is set, its value will be used instead.",
    envvar="ARGOS_YAML_FILE",
    callback=validate_config_access,
)
@click.option("--name", required=True, help="Name of the user to delete.")
@coroutine
async def delete(config, name):
    """Delete user"""
    # It’s mandatory to do it before the imports
    os.environ["ARGOS_YAML_FILE"] = config

    # The imports are made here otherwise the agent will need server configuration files.
    from argos.server import queries

    db = await get_db()
    _user = await queries.get_user(db, name)
    if _user is None:
        click.echo(f"User {name} does not exist.")
        sysexit(1)

    db.delete(_user)
    db.commit()

    click.echo(f"User {name} deleted.")


@user.command()
@click.option(
    "--config",
    default="argos-config.yaml",
    help="Path of the configuration file. "
    "If ARGOS_YAML_FILE environment variable is set, its value will be used instead.",
    envvar="ARGOS_YAML_FILE",
    callback=validate_config_access,
)
@coroutine
async def show(config):
    """List all users"""
    # It’s mandatory to do it before the imports
    os.environ["ARGOS_YAML_FILE"] = config

    # The imports are made here otherwise the agent will need server configuration files.
    from argos.server import queries

    db = await get_db()
    users = await queries.list_users(db)
    if users.count() == 0:
        click.echo("There is no users in database.")
        sysexit(1)

    click.echo("✅ means that the user is enabled.")
    click.echo("❌ means that the user is disabled.")

    for _user in users.all():
        status = "✅"
        if _user.disabled:
            status = "❌"
        click.echo(f"{status} {_user.username}, last login: {_user.last_login_at}")


@server.command(short_help="Generate a token for agents")
@coroutine
async def generate_token():
    """Generate a token, which can be used as an agent’s authentication token.

    It’s actually an UUID
    """
    click.echo(uuid4())


@server.command()
@coroutine
async def generate_config():
    """Output a self-documented example config file.

    \b
    Redirect the output to a file to save it:
        argos server generate-config > /etc/argos/config.yaml
    """
    config_example = Path(__file__).resolve().parent / "config-example.yaml"
    with config_example.open("r", encoding="utf-8") as f:
        print(f.read())


@server.command()
@click.option(
    "--config",
    default="argos-config.yaml",
    help="Path of the configuration file. "
    "If ARGOS_YAML_FILE environment variable is set, its value will be used instead.",
    envvar="ARGOS_YAML_FILE",
    callback=validate_config_access,
)
@click.option("--domain", help="Domain for the notification", default="example.org")
@click.option("--severity", help="Severity", default="CRITICAL")
@coroutine
async def test_mail(config, domain, severity):
    """Send a test email"""
    os.environ["ARGOS_YAML_FILE"] = config

    from datetime import datetime

    from argos.logging import set_log_level
    from argos.server.alerting import notify_by_mail
    from argos.server.main import read_config
    from argos.server.models import Result, Task

    conf = read_config(config)

    if not conf.general.mail:
        click.echo("Mail is not configured, cannot test", err=True)
        sysexit(1)
    else:
        now = datetime.now()
        task = Task(
            url=f"https://{domain}",
            domain=domain,
            check="body-contains",
            expected="foo",
            frequency=1,
            selected_by="test",
            selected_at=now,
        )

        result = Result(
            submitted_at=now,
            status="success",
            context={"foo": "bar"},
            task=task,
            agent_id="test",
            severity="ok",
        )

        class _FalseRequest:
            def url_for(*args, **kwargs):
                return "/url"

        set_log_level("debug")
        notify_by_mail(
            result,
            task,
            severity=severity,
            old_severity="OLD SEVERITY",
            config=conf.general.mail,
            request=_FalseRequest(),
        )


@server.command()
@click.option(
    "--config",
    default="argos-config.yaml",
    help="Path of the configuration file. "
    "If ARGOS_YAML_FILE environment variable is set, its value will be used instead.",
    envvar="ARGOS_YAML_FILE",
    callback=validate_config_access,
)
@click.option("--domain", help="Domain for the notification", default="example.org")
@click.option("--severity", help="Severity", default="CRITICAL")
@coroutine
async def test_gotify(config, domain, severity):
    """Send a test gotify notification"""
    os.environ["ARGOS_YAML_FILE"] = config

    from datetime import datetime

    from argos.logging import set_log_level
    from argos.server.alerting import notify_with_gotify
    from argos.server.main import read_config
    from argos.server.models import Result, Task

    conf = read_config(config)

    if not conf.general.gotify:
        click.echo("Gotify notifications are not configured, cannot test", err=True)
        sysexit(1)
    else:
        now = datetime.now()
        task = Task(
            url=f"https://{domain}",
            domain=domain,
            check="body-contains",
            expected="foo",
            frequency=1,
            selected_by="test",
            selected_at=now,
        )

        result = Result(
            submitted_at=now,
            status="success",
            context={"foo": "bar"},
            task=task,
            agent_id="test",
            severity="ok",
        )

        class _FalseRequest:
            def url_for(*args, **kwargs):
                return "/url"

        set_log_level("debug")
        notify_with_gotify(
            result,
            task,
            severity=severity,
            old_severity="OLD SEVERITY",
            config=conf.general.gotify,
            request=_FalseRequest(),
        )


@server.command()
@click.option(
    "--config",
    default="argos-config.yaml",
    help="Path of the configuration file. "
    "If ARGOS_YAML_FILE environment variable is set, its value will be used instead.",
    envvar="ARGOS_YAML_FILE",
    callback=validate_config_access,
)
@click.option("--domain", help="Domain for the notification", default="example.org")
@click.option("--severity", help="Severity", default="CRITICAL")
@click.option(
    "--apprise-group", help="Apprise group for the notification", required=True
)
@coroutine
async def test_apprise(config, domain, severity, apprise_group):
    """Send a test apprise notification"""
    os.environ["ARGOS_YAML_FILE"] = config

    from datetime import datetime

    from argos.logging import set_log_level
    from argos.server.alerting import notify_with_apprise
    from argos.server.main import read_config
    from argos.server.models import Result, Task

    conf = read_config(config)

    if not conf.general.apprise:
        click.echo("Apprise notifications are not configured, cannot test", err=True)
        sysexit(1)
    else:
        now = datetime.now()
        task = Task(
            url=f"https://{domain}",
            domain=domain,
            check="body-contains",
            expected="foo",
            frequency=1,
            selected_by="test",
            selected_at=now,
        )

        result = Result(
            submitted_at=now,
            status="success",
            context={"foo": "bar"},
            task=task,
            agent_id="test",
            severity="ok",
        )

        class _FalseRequest:
            def url_for(*args, **kwargs):
                return "/url"

        set_log_level("debug")
        notify_with_apprise(
            result,
            task,
            severity=severity,
            old_severity="OLD SEVERITY",
            group=conf.general.apprise[apprise_group],
            request=_FalseRequest(),
        )


@server.command(short_help="Nagios compatible severities report")
@click.option(
    "--config",
    default="argos-config.yaml",
    help="Path of the configuration file. "
    "If ARGOS_YAML_FILE environment variable is set, its value will be used instead.",
    envvar="ARGOS_YAML_FILE",
    callback=validate_config_access,
)
@coroutine
async def nagios(config):
    """Output a report of current severities suitable for Nagios
    with a Nagios compatible exit code"""
    os.environ["ARGOS_YAML_FILE"] = config

    # The imports are made here otherwise the agent will need server configuration files.
    from argos.server import queries

    exit_nb = 0
    db = await get_db()
    severities = await queries.get_severity_counts(db)

    if severities["warning"] != 0:
        exit_nb = 1
    if severities["critical"] != 0:
        exit_nb = 2
    if severities["unknown"] != 0:
        exit_nb = 2

    stats = (
        f"ok={severities['ok']}; warning={severities['warning']}; "
        f"critical={severities['critical']}; unknown={severities['unknown']};"
    )

    if exit_nb == 0:
        print("OK — All sites are ok|{stats}")
    elif exit_nb == 1:
        print(f"WARNING — {severities['warning']} sites are in warning state|{stats}")
    elif severities["critical"] == 0:
        print(f"UNKNOWN — {severities['unknown']} sites are in unknown state|{stats}")
    elif severities["unknown"] == 0:
        print(
            f"CRITICAL — {severities['critical']} sites are in critical state|{stats}"
        )
    else:
        print(
            f"CRITICAL/UNKNOWN — {severities['critical']} sites are in critical state "
            f"and {severities['unknown']} sites are in unknown state|{stats}"
        )

    sysexit(exit_nb)


if __name__ == "__main__":
    cli()



"""Pydantic schemas for configuration

For database models, see argos.server.models.
"""

import json

from typing import Any, Dict, List, Literal, Tuple

from durations_nlp import Duration
from pydantic import (
    BaseModel,
    ConfigDict,
    HttpUrl,
    PostgresDsn,
    StrictBool,
    EmailStr,
    PositiveInt,
    field_validator,
)
from pydantic.functional_validators import AfterValidator, BeforeValidator
from pydantic.networks import UrlConstraints
from pydantic_core import Url
from typing_extensions import Annotated

from argos.schemas.utils import Method

Severity = Literal["warning", "error", "critical", "unknown"]
Environment = Literal["dev", "test", "production"]
Unauthenticated = Literal["dashboard", "all"]
SQLiteDsn = Annotated[
    Url,
    UrlConstraints(
        allowed_schemes=["sqlite"],
    ),
]


def parse_threshold(value):
    """Parse duration threshold for SSL certificate validity"""
    for duration_str, severity in value.items():
        days = Duration(duration_str).to_days()
        # Return here because it's one-item dicts.
        return (days, severity)


class SSL(BaseModel):
    thresholds: List[Annotated[Tuple[int, Severity], BeforeValidator(parse_threshold)]]


class WebsiteCheck(BaseModel):
    key: str
    value: str | List[str] | Dict[str, str]
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        if isinstance(value, str):
            return {"expected": value}
        if isinstance(value, dict):
            return value
        if isinstance(value, list):
            return {"expected": value}

        raise ValueError("Invalid type")


def parse_checks(value):
    """Check that checks are valid (i.e. registered) checks"""

    # To avoid circular imports
    from argos.checks import get_registered_checks

    available_names = get_registered_checks().keys()

    for name, expected in value.items():
        if name not in available_names:
            msg = f"Check should be one of f{available_names}. ({name} given)"
            raise ValueError(msg)
        if name == "http-to-https":
            if isinstance(expected, int) and expected in range(300, 400):
                expected = json.dumps({"value": expected})
            elif isinstance(expected, list):
                expected = json.dumps({"list": expected})
            elif (
                isinstance(expected, dict)
                and "start" in expected
                and "stop" in expected
            ):
                expected = json.dumps({"range": [expected["start"], expected["stop"]]})
            else:
                expected = json.dumps({"range": [300, 400]})
        else:
            if isinstance(expected, int):
                expected = str(expected)
            if isinstance(expected, list):
                expected = json.dumps(expected)
            if isinstance(expected, dict):
                expected = json.dumps(expected)
        return (name, expected)


def parse_request_data(value):
    """Turn form or JSON data into JSON string"""

    return json.dumps(
        {"data": value.data, "json": value.is_json, "headers": value.headers}
    )


class RequestData(BaseModel):
    data: Any = None
    is_json: bool = False
    headers: Dict[str, str] | None = None


class WebsitePath(BaseModel):
    path: str
    method: Method = "GET"
    request_data: Annotated[
        RequestData, AfterValidator(parse_request_data)
    ] | None = None
    checks: List[
        Annotated[
            Tuple[str, str],
            BeforeValidator(parse_checks),
        ]
    ]


class Website(BaseModel):
    domain: HttpUrl
    ipv4: bool | None = None
    ipv6: bool | None = None
    frequency: float | None = None
    recheck_delay: float | None = None
    retry_before_notification: int | None = None
    paths: List[WebsitePath]

    @field_validator("frequency", mode="before")
    def parse_frequency(cls, value):
        """Convert the configured frequency to minutes"""
        if value:
            return Duration(value).to_minutes()

        return None

    @field_validator("recheck_delay", mode="before")
    def parse_recheck_delay(cls, value):
        """Convert the configured recheck delay to minutes"""
        if value:
            return Duration(value).to_minutes()

        return None


class Service(BaseModel):
    """List of agents’ token"""

    secrets: List[str]


class MailAuth(BaseModel):
    """Mail authentication configuration"""

    login: str
    password: str


class Mail(BaseModel):
    """Mail configuration"""

    mailfrom: EmailStr
    host: str = "127.0.0.1"
    port: PositiveInt = 25
    ssl: StrictBool = False
    starttls: StrictBool = False
    auth: MailAuth | None = None
    addresses: List[EmailStr]


class Alert(BaseModel):
    """List of way to handle alerts, by severity"""

    ok: List[str]
    warning: List[str]
    critical: List[str]
    unknown: List[str]


class GotifyUrl(BaseModel):
    url: HttpUrl
    tokens: List[str]


class DbSettings(BaseModel):
    url: PostgresDsn | SQLiteDsn
    pool_size: int = 10
    max_overflow: int = 20


class LdapSettings(BaseModel):
    uri: str
    user_tree: str
    bind_dn: str | None = None
    bind_pwd: str | None = None
    user_attr: str
    user_filter: str | None = None


class General(BaseModel):
    """Frequency for the checks and alerts"""

    db: DbSettings
    env: Environment = "production"
    cookie_secret: str
    session_duration: int = 10080  # 7 days
    remember_me_duration: int | None = None
    unauthenticated_access: Unauthenticated | None = None
    ldap: LdapSettings | None = None
    frequency: float
    recheck_delay: float | None = None
    retry_before_notification: int = 0
    ipv4: bool = True
    ipv6: bool = True
    root_path: str = ""
    alerts: Alert
    mail: Mail | None = None
    gotify: List[GotifyUrl] | None = None
    apprise: Dict[str, List[str]] | None = None

    @field_validator("session_duration", mode="before")
    def parse_session_duration(cls, value):
        """Convert the configured session duration to minutes"""
        return Duration(value).to_minutes()

    @field_validator("remember_me_duration", mode="before")
    def parse_remember_me_duration(cls, value):
        """Convert the configured session duration with remember me feature to minutes"""
        if value:
            return int(Duration(value).to_minutes())

        return None

    @field_validator("frequency", mode="before")
    def parse_frequency(cls, value):
        """Convert the configured frequency to minutes"""
        return Duration(value).to_minutes()

    @field_validator("recheck_delay", mode="before")
    def parse_recheck_delay(cls, value):
        """Convert the configured recheck delay to minutes"""
        if value:
            return Duration(value).to_minutes()

        return None


class Config(BaseModel):
    general: General
    service: Service
    ssl: SSL
    websites: List[Website]
